// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Windows.Forms;

using Smdn.Collections;
using Smdn.Formats.Thbgm;

namespace Smdn.Windows.Forms.ThbgmPlayer {
  internal interface ITrackListView {
    void UpdateAllTracksItemSubItems();
  }

  public class TrackListView : TrackListView<TrackListViewItem> {
  }

  public class TrackListView<TTrackListViewItem> : ListView, ITrackListView where TTrackListViewItem : TrackListViewItem, new() {
    public event CancelEventHandler ProductChanging;
    public event EventHandler ProductChanged;

    public ProductInfo Product {
      get { return product; }
      set
      {
        if (product == value)
          return;

        var e = new CancelEventArgs();

        OnProductChanging(e);

        if (e.Cancel)
          return;

        product = value;

        OnProductChanged();
      }
    }

    public Track SelectedTrack {
      get
      {
        if (SelectedItems.Count <= 0)
          return null;
        else
          return (SelectedItems[0] as TrackListViewItem).Track;
      }
    }

    public IEnumerable<Track> Tracks {
      get { return tracks; }
    }

    public IEnumerable<TTrackListViewItem> TrackItems {
      get
      {
        foreach (ListViewItem item in Items) {
          if (item == allTracksItem)
            continue;
          else if (item is TTrackListViewItem)
            yield return item as TTrackListViewItem;
          else
            continue;
        }
      }
    }

    protected ListViewItem AllTracksItem {
      get { return allTracksItem; }
    }

    public TrackListView()
      : base()
    {
      CheckBoxes = true;
      View = View.Details;
      HeaderStyle = ColumnHeaderStyle.Nonclickable;
      MultiSelect = false;
      FullRowSelect = true;
      HideSelection = false;

      Columns.Add(Texts.GetText("TrackNumber"), 45);
      Columns.Add(Texts.GetText("Title"), 260);
#if DEBUG
      Columns.Add("IsDirty", 20);
#endif
      Columns.Add(Texts.GetText("TotalLength"), 100);
      Columns.Add(Texts.GetText("RepeatedTimes"), 70);
      Columns.Add(Texts.GetText("FadingStart"), 100);
      Columns.Add(Texts.GetText("FadingLength"), 140);
      Columns.Add(Texts.GetText("FadingCurveType"), 120);
      Columns.Add(Texts.GetText("FadingCurveCoefficient"), 45);
      Columns.Add(Texts.GetText("TrailingSilenceLength"), 100);

      allTracksItem = new ListViewItem();
    }

    public virtual void UpdateAllTracksItemSubItems()
    {
      allTracksItem.SubItems.Clear();

      allTracksItem.Text = Texts.GetText("AllTracks");
      allTracksItem.SubItems.Add(product == null ? string.Empty : product.Title);
#if DEBUG
      allTracksItem.SubItems.Add("-"); // IsDirty
#endif

      var totalLength = TimeSpan.Zero;

      foreach (var item in TrackItems) {
        if (item.Track != null)
          totalLength += item.Track.Effect.TotalLength;
      }

      allTracksItem.SubItems.Add(TrackListViewItem.ConvertTimeSpanToString(totalLength));

      // XXX
      var firstItem = TrackItems.FirstOrDefault();
      var firstTrack = firstItem == null ? null : firstItem.Track;
      var subItemValues = (firstTrack == null)
        ? new List<string>() {
          string.Empty,
          string.Empty,
          string.Empty,
          string.Empty,
          string.Empty,
          string.Empty,
        }
        : new List<string>() {
          firstTrack.TrackInfo.CalculateRepeatedTimes(firstTrack.Effect.TotalLength).ToString(),
          TrackListViewItem.ConvertTimeSpanToString(firstTrack.Effect.FadingStart),
          TrackListViewItem.ConvertTimeSpanToString(firstTrack.Effect.FadingLength),
          Texts.GetText(firstTrack.Effect.FadingCurveType),
          firstTrack.Effect.FadingCurveCoefficient.ToString("F2"),
          TrackListViewItem.ConvertTimeSpanToString(firstTrack.Effect.TrailingSilenceLength),
        };

      var multipleValueString = Texts.GetText("MultipleValues");

      if (firstTrack != null) {
        foreach (var item in TrackItems) {
          var track = item.Track;

          if (track == null)
            continue;

          if (firstTrack.TrackInfo.CalculateRepeatedTimes(firstTrack.Effect.TotalLength) != track.TrackInfo.CalculateRepeatedTimes(item.Track.Effect.TotalLength))
            subItemValues[0] = multipleValueString;
          if (firstTrack.Effect.FadingStart != track.Effect.FadingStart)
            subItemValues[1] = multipleValueString;
          if (firstTrack.Effect.FadingLength != track.Effect.FadingLength)
            subItemValues[2] = multipleValueString;
          if (firstTrack.Effect.FadingCurveType != track.Effect.FadingCurveType)
            subItemValues[3] = multipleValueString;
          if (firstTrack.Effect.FadingCurveCoefficient != track.Effect.FadingCurveCoefficient)
            subItemValues[4] = multipleValueString;
          if (firstTrack.Effect.TrailingSilenceLength != track.Effect.TrailingSilenceLength)
            subItemValues[5] = multipleValueString;
        }
      }

      allTracksItem.SubItems.AddRange(subItemValues.ToArray());
    }

    public void SetProductAndTracks(ProductInfo product, IEnumerable<Track> tracks)
    {
      SetTracks(tracks);

      if (this.product != product) {
        this.product = product;

        // raise event
        OnProductChanged(EventArgs.Empty);
      }
    }

    public void CheckAll()
    {
      allTracksItem.Checked = true;
    }

    public void UncheckAll()
    {
      allTracksItem.Checked = false;
    }

    private void SetTracks(IEnumerable<Track> tracks)
    {
      this.tracks.Clear();

      itemsUpdating = true;

      BeginUpdate();

      Items.Clear();

      foreach (var track in tracks) {
        var item = new TTrackListViewItem();

        Items.Add(item);

        item.Track = track;
        item.Checked = track.Checked;

        this.tracks.Add(track);
      }

      Items.Add(allTracksItem);

      UpdateAllTracksItemSubItems();

      EndUpdate();

      itemsUpdating = false;
    }

    protected override void OnItemChecked(ItemCheckedEventArgs e)
    {
      if (e.Item == allTracksItem) {
        // check/uncheck all
        foreach (var item in TrackItems) {
          item.Checked = allTracksItem.Checked;
        }
      }
      else {
        var item = (e.Item as TTrackListViewItem);

        if (item.Track != null) {
          item.Track.Checked = item.Checked;

          if (!itemsUpdating)
            // edit by user operation
            item.Track.SetDirty();
        }
      }

      base.OnItemChecked(e);
    }

    protected override void OnItemSelectionChanged(ListViewItemSelectionChangedEventArgs e)
    {
      if (e.Item == allTracksItem && e.IsSelected)
        e.Item.Selected = false;

      base.OnItemSelectionChanged(e);
    }

    private void OnProductChanged()
    {
      if (product == null)
        SetTracks(new Track[] {});
      else
        SetTracks(product.Tracks.ConvertAll(delegate(TrackInfo trackInfo) {
          return new Track(trackInfo);
        }));

      allTracksItem.Checked = true;

      OnProductChanged(EventArgs.Empty);
    }

    protected virtual void OnProductChanging(CancelEventArgs e)
    {
      var ev = this.ProductChanging;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnProductChanged(EventArgs e)
    {
      var ev = this.ProductChanged;

      if (ev != null)
        ev(this, e);
    }

    private ProductInfo product = null;
    private List<Track> tracks = new List<Track>();
    private bool itemsUpdating = false;
    private ListViewItem allTracksItem;
  }
}
