using System;
using System.Collections.Generic;
using System.IO;

using Smdn.Collections;
using Smdn.Formats;
using Smdn.Formats.Ini;
using Smdn.Formats.Playlists;
using Smdn.Formats.Thbgm;
using Smdn.Windows.Forms.ThbgmPlayer;

namespace Smdn.Applications.ThbgmExtractor {
  public static class Config {
    internal const string DefaultConfigFileName = "ThbgmExtractor.ini";

    public static class RecentFiles {
      private static string SectionName = "RecentFiles";
      private const int MaxEntries = 10;

      private static List<string> ExtractionConfigFiles = new List<string>();

      public static void ClearExtractionConfigFiles()
      {
        lock (((System.Collections.ICollection)ExtractionConfigFiles).SyncRoot) {
          ExtractionConfigFiles.Clear();
        }
      }

      public static void AddExtractionConfigFile(string file)
      {
        AddExtractionConfigFile(file, true);
      }

      private static void AddExtractionConfigFile(string file, bool insert)
      {
        lock (((System.Collections.ICollection)ExtractionConfigFiles).SyncRoot) {
          if (ExtractionConfigFiles.Contains(file))
            ExtractionConfigFiles.Remove(file);
          else if (MaxEntries <= ExtractionConfigFiles.Count)
            ExtractionConfigFiles.RemoveAt(ExtractionConfigFiles.Count - 1);

          if (insert)
            ExtractionConfigFiles.Insert(0, file);
          else
            ExtractionConfigFiles.Add(file);
        }
      }

      public static string[] GetExtractionConfigFiles()
      {
        lock (((System.Collections.ICollection)ExtractionConfigFiles).SyncRoot) {
          return ExtractionConfigFiles.ToArray();
        }
      }

      internal static void LoadFrom(IniDocument document)
      {
        var section = document[SectionName];
        var entries = new List<KeyValuePair<string, string>>(section.Entries);

        entries.Sort(delegate(KeyValuePair<string, string> x, KeyValuePair<string, string> y) {
          return string.Compare(x.Key, y.Key);
        });

        foreach (var entry in entries) {
          if (entry.Key.StartsWith("ExtractionConfigFile")) {
            if (File.Exists(entry.Value))
              AddExtractionConfigFile(entry.Value, false);
          }
        }
      }

      internal static void SaveTo(IniDocument document)
      {
        var section = document[SectionName];

        var extractionConfigFiles = GetExtractionConfigFiles();

        for (var i = 0; i < extractionConfigFiles.Length; i++) {
          section[string.Format("ExtractionConfigFile{0}", i)] = extractionConfigFiles[i];
        }
      }
    }

    public static class ExtractionConfigDefaults {
      private static string SectionName = "ExtractionConfigDefaults";

      public static bool SaveSourceAfterEncode = false;
      public static PlaylistFormat PlaylistFormat = null;

      internal static void LoadFrom(IniDocument document)
      {
        var section = document[SectionName];

        SaveSourceAfterEncode = section.Get<bool>("SaveSourceAfterEncode", SaveSourceAfterEncode, Convert.ToBoolean);

        var playlistFormatName = section["PlaylistFormat"];

        PlaylistFormat = PlaylistFormat.Formats.Find(delegate(PlaylistFormat format) {
          return format.Name == playlistFormatName;
        }) ?? PlaylistFormat;
      }

      internal static void SaveTo(IniDocument document)
      {
        var section = document[SectionName];

        section["SaveSourceAfterEncode"] = SaveSourceAfterEncode.ToString();
        section["PlaylistFormat"] = (PlaylistFormat == null) ? string.Empty : PlaylistFormat.Name;
      }
    }

    public static class CollectiveEffectEditDefaults {
      private static string SectionName = "CollectiveEffectEditDefaults";

      public static EffectLengthBasis EffectLengthBasis = EffectLengthBasis.FadingStart;
      public static FadingCurveType FadingCurveType = FadingCurveType.Linear;
      public static double FadingCurveCoefficient = 1.0;
      public static int TimesToRepeat = 2; // 2 loops
      public static TimeSpan FadingStart = TimeSpan.FromMinutes(5.0); // 5 mins
      public static TimeSpan FadingOffset = TimeSpan.Zero; // 0 secs
      public static TimeSpan FadingLength = TimeSpan.FromSeconds(20.0); // 20 secs
      public static TimeSpan TotalLength = TimeSpan.FromMinutes(5.0); // 5 mins
      public static TimeSpan TrailingSilenceLength = TimeSpan.Zero; // 0 secs

      internal static void LoadFrom(IniDocument document)
      {
        var section = document[SectionName];

        EffectLengthBasis       = section.Get<EffectLengthBasis>("EffectLengthBasis", EffectLengthBasis, EnumUtils.Parse<EffectLengthBasis>);
        FadingCurveType         = section.Get<FadingCurveType>("FadingCurveType", FadingCurveType, EnumUtils.Parse<FadingCurveType>);
        FadingCurveCoefficient  = section.Get<double>("FadingCurveCoefficient", FadingCurveCoefficient, Convert.ToDouble);
        TimesToRepeat           = section.Get<int>("TimesToRepeat", TimesToRepeat, Convert.ToInt32);
        FadingStart             = TimeSpanUtils.ParsePositive(section["FadingStart"], FadingStart);
        FadingOffset            = TimeSpanUtils.Parse(section["FadingOffset"], FadingOffset);
        FadingLength            = TimeSpanUtils.ParsePositive(section["FadingLength"], FadingLength);
        TotalLength             = TimeSpanUtils.ParsePositive(section["TotalLength"], TotalLength);
        TrailingSilenceLength   = TimeSpanUtils.ParsePositive(section["TrailingSilenceLength"], TrailingSilenceLength);

        if (double.IsNaN(FadingCurveCoefficient) || double.IsInfinity(FadingCurveCoefficient))
          FadingCurveCoefficient = 1.0;

        if (TimesToRepeat < 0)
          TimesToRepeat = 0;
      }

      internal static void SaveTo(IniDocument document)
      {
        var section = document[SectionName];

        section["EffectLengthBasis"]      = EffectLengthBasis.ToString();
        section["FadingCurveType"]        = FadingCurveType.ToString();
        section["FadingCurveCoefficient"] = FadingCurveCoefficient.ToString();
        section["TimesToRepeat"]          = TimesToRepeat.ToString();
        section["FadingStart"]            = FadingStart.ToString();
        section["FadingOffset"]           = FadingOffset.ToString();
        section["FadingLength"]           = FadingLength.ToString();
        section["TotalLength"]            = TotalLength.ToString();
        section["TrailingSilenceLength"]  = TrailingSilenceLength.ToString();
      }

      internal static void OpenConfigAppliedDialog(Action<CollectiveEffectEditDialog> action)
      {
        using (var dialog = new CollectiveEffectEditDialog()) {
          Apply(dialog);

          action(dialog);

          if (dialog.DialogResult == System.Windows.Forms.DialogResult.OK)
            Store(dialog);
        }
      }

      internal static void Apply(CollectiveEffectEditDialog dialog)
      {
        dialog.EffectLengthBasis      = EffectLengthBasis;
        dialog.TimesToRepeat          = TimesToRepeat;
        dialog.FadingStart            = FadingStart;
        dialog.FadingOffset           = FadingOffset;
        dialog.FadingLength           = FadingLength;
        dialog.TotalLength            = TotalLength;
        dialog.TrailingSilenceLength  = TrailingSilenceLength;

        dialog.SetFadingCurveValues(FadingCurveType, FadingCurveCoefficient);
      }

      internal static void Store(CollectiveEffectEditDialog dialog)
      {
        EffectLengthBasis      = dialog.EffectLengthBasis;
        FadingCurveType        = dialog.FadingCurveType;
        FadingCurveCoefficient = dialog.FadingCurveCoefficient;
        TimesToRepeat          = dialog.TimesToRepeat;
        FadingStart            = dialog.FadingStart;
        FadingOffset           = dialog.FadingOffset;
        FadingLength           = dialog.FadingLength;
        TotalLength            = dialog.TotalLength;
        TrailingSilenceLength  = dialog.TrailingSilenceLength;
      }
    }

    public static class TrackFormatInfoDefaults {
      private static string SectionName = "TrackFormatInfoDefaults";

      public static string Format = "$trackno. $title";
      public static TrackFormatInfo TrackFormatInfo;

      static TrackFormatInfoDefaults()
      {
        TrackFormatInfo = new TrackFormatInfo();
        TrackFormatInfo.ZeroPadding = true;
        TrackFormatInfo.TitleUnificationOptions = TitleUnificationOptions.FullWidthTilde | TitleUnificationOptions.SingleIdeographicSpace;
      }

      internal static void LoadFrom(IniDocument document)
      {
        var section = document[SectionName];

        Format = section.Get("Format", Format);
        TrackFormatInfo.TitleUnificationOptions = section.Get<TitleUnificationOptions>("TitleUnificationOptions", TrackFormatInfo.TitleUnificationOptions, EnumUtils.Parse<TitleUnificationOptions>);
        TrackFormatInfo.ZeroPadding = section.Get<bool>("ZeroPadding", TrackFormatInfo.ZeroPadding, Convert.ToBoolean);
      }

      internal static void SaveTo(IniDocument document)
      {
        var section = document[SectionName];

        section["Format"] = Format;
        section["TitleUnificationOptions"] = TrackFormatInfo.TitleUnificationOptions.ToString();
        section["ZeroPadding"] = TrackFormatInfo.ZeroPadding.ToString();
      }

      internal static void OpenConfigAppliedFileNameEditDialog(Action<FileNameEditDialog> action)
      {
        using (var dialog = new FileNameEditDialog()) {
          Apply(dialog);

          action(dialog);

          if (dialog.DialogResult == System.Windows.Forms.DialogResult.OK)
            Store(dialog);
        }
      }

      internal static void Apply(FileNameEditDialog dialog)
      {
        dialog.Format = Format;
        dialog.TrackFormatInfo = TrackFormatInfo.Clone();
      }

      internal static void Store(FileNameEditDialog dialog)
      {
        Format = dialog.Format;

        if (dialog.TrackFormatInfo == null)
          TrackFormatInfo = new TrackFormatInfo();
        else
          TrackFormatInfo = dialog.TrackFormatInfo.Clone();
      }
    }

    public static class UserEnvironmentProductInfoes {
      private const string sectionPrefix = "ProductInfo:";

      private static string GetSectionNameFrom(ProductInfo product)
      {
        if (string.IsNullOrEmpty(product.Title))
          // can't store, ignore
          return null;
        else
          return sectionPrefix + product.Title;
      }

      private static string GetProductTitleFrom(string sectionName)
      {
        if (sectionName.StartsWith(sectionPrefix))
          return sectionName.Substring(sectionPrefix.Length);
        else
          return null;
      }

      internal static void LoadFrom(IniDocument document, IEnumerable<ProductInfo> products)
      {
        var productList = new List<ProductInfo>(products);

        foreach (var section in document) {
          var title = GetProductTitleFrom(section.Name);
          var product = productList.Find(delegate(ProductInfo p) {
            return p.Title == title;
          });

          if (product == null)
            continue;

          var installedBgmSourcePath = section["InstalledBgmSourcePath"];

          if (File.Exists(installedBgmSourcePath))
            product.InstalledBgmSourcePath = installedBgmSourcePath;
        }
      }

      internal static void SaveTo(IniDocument document, IEnumerable<ProductInfo> products)
      {
        foreach (var product in products) {
          var sectionName = GetSectionNameFrom(product);

          if (sectionName == null)
            continue;

          var section = document[sectionName];

          if (!string.IsNullOrEmpty(product.InstalledBgmSourcePath))
            section["InstalledBgmSourcePath"] = product.InstalledBgmSourcePath;

          // delete empty section
          if (section.Entries.Count == 0)
            document.Remove(section.Name);
        }
      }
    }

    public static void Load(string searchDirectory, IEnumerable<ProductInfo> products)
    {
      configFilePath = Path.Combine(searchDirectory, DefaultConfigFileName);

      if (!File.Exists(configFilePath)) {
        Console.WriteLine("configuration file '{0}' not found", configFilePath);
        return;
      }

      try {
        var document = IniDocument.Load(configFilePath, StringComparer.CurrentCulture);

        RecentFiles.LoadFrom(document);
        ExtractionConfigDefaults.LoadFrom(document);
        CollectiveEffectEditDefaults.LoadFrom(document);
        TrackFormatInfoDefaults.LoadFrom(document);
        UserEnvironmentProductInfoes.LoadFrom(document, products);
      }
      catch {
        // ignore any exceptions
      }
    }

    public static void Save(IEnumerable<ProductInfo> products)
    {
      try {
        var document = new IniDocument(StringComparer.CurrentCulture);

        RecentFiles.SaveTo(document);
        ExtractionConfigDefaults.SaveTo(document);
        CollectiveEffectEditDefaults.SaveTo(document);
        TrackFormatInfoDefaults.SaveTo(document);
        UserEnvironmentProductInfoes.SaveTo(document, products);

        document.Save(configFilePath);
      }
      catch {
        // ignore any exceptions
      }
    }

    private static string configFilePath;
  }
}
