using System;
using System.IO;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;
using System.Xml;

using Smdn.Collections;
using Smdn.Formats.Ini;
using Smdn.Formats.Playlists;
using Smdn.Formats.Thbgm;
using Smdn.IO;
using Smdn.Windows.Forms.ThbgmPlayer;
using Smdn.Xml;

namespace Smdn.Applications.ThbgmExtractor {
  public class ExtractionConfig : ICloneable, INotifyPropertyChanging, INotifyPropertyChanged {
    public static string FileExtension = ".ini";
    public static string THxxBGMFadeoutSettingsFileExtension = ".fsx";

    public event PropertyChangingEventHandler PropertyChanging;
    public event PropertyChangedEventHandler PropertyChanged;

    private ProductInfo product;

    public ProductInfo Product {
      get { return product; }
      set
      {
        if (product == value)
          return;

        if (OnPropertyChanging("Product")) {
          product = value;

          OnPropertyChanged("Product");
        }
      }
    }

    private string bgmSourcePath;

    public string BgmSourcePath {
      get { return bgmSourcePath; }
      set
      {
        if (bgmSourcePath == value)
          return;

        if (OnPropertyChanging("BgmSourcePath")) {
          bgmSourcePath = value;

          OnPropertyChanged("BgmSourcePath");
        }
      }
    }

    private string outputDirectory;

    public string OutputDirectory {
      get { return outputDirectory; }
      set
      {
        if (outputDirectory == value)
          return;

        if (OnPropertyChanging("OutputDirectory")) {
          outputDirectory = value;

          OnPropertyChanged("OutputDirectory");
        }
      }
    }

    private TrackFormatInfo trackFormatInfo = Config.TrackFormatInfoDefaults.TrackFormatInfo.Clone();

    public TrackFormatInfo TrackFormatInfo {
      get { return trackFormatInfo; }
      set
      {
        if (trackFormatInfo == value)
          return;

        if (OnPropertyChanging("TrackFormatInfo")) {
          trackFormatInfo = value;

          OnPropertyChanged("TrackFormatInfo");
        }
      }
    }

    private string trackFormatString = Config.TrackFormatInfoDefaults.Format;

    public string TrackFormatString {
      get { return trackFormatString; }
      set
      {
        if (trackFormatString == value)
          return;

        if (OnPropertyChanging("TrackFormatString")) {
          trackFormatString = value;

          OnPropertyChanged("TrackFormatString");
        }
      }
    }

    private EncoderConfig encoderConfig = EncoderConfig.Empty;

    public EncoderConfig EncoderConfig {
      get { return encoderConfig; }
      set
      {
        if (encoderConfig == value)
          return;

        if (OnPropertyChanging("EncoderConfig")) {
          encoderConfig = value;

          OnPropertyChanged("EncoderConfig");
        }
      }
    }

    private bool saveSourceAfterEncode = Config.ExtractionConfigDefaults.SaveSourceAfterEncode;

    public bool SaveSourceAfterEncode {
      get { return saveSourceAfterEncode; }
      set
      {
        if (saveSourceAfterEncode == value)
          return;

        if (OnPropertyChanging("SaveSourceAfterEncode")) {
          saveSourceAfterEncode = value;

          OnPropertyChanged("SaveSourceAfterEncode");
        }
      }
    }

    private PlaylistFormat playlistFormat = Config.ExtractionConfigDefaults.PlaylistFormat;

    public PlaylistFormat PlaylistFormat {
      get { return playlistFormat; }
      set
      {
        if (playlistFormat == value)
          return;

        if (OnPropertyChanging("PlaylistFormat")) {
          playlistFormat = value;

          OnPropertyChanged("PlaylistFormat");
        }
      }
    }

    private IEnumerable<Track> tracks = new Track[] {};

    public IEnumerable<Track> Tracks {
      get { return tracks; }
      set
      {
        if (tracks == value)
          return;

        if (OnPropertyChanging("Tracks")) {
          tracks = value;

          OnPropertyChanged("Tracks");
        }
      }
    }

    private Dictionary<Track, string> extractedFilePaths = new Dictionary<Track, string>();

    public Dictionary<Track, string> ExtractedFilePaths {
      get { return extractedFilePaths; }
    }

    private Dictionary<Track, string> encodedFilePaths = new Dictionary<Track, string>();

    public Dictionary<Track, string> EncodedFilePaths {
      get { return encodedFilePaths; }
    }

    public bool AreAnyTracksDirty {
      get
      {
        if (tracks == null)
          return false;

        foreach (var track in tracks) {
          if (track.IsDirty)
            return true;
        }

        return false;
      }
    }

    object ICloneable.Clone()
    {
      return Clone();
    }

    public virtual ExtractionConfig Clone()
    {
      var cloned = (ExtractionConfig)MemberwiseClone();

      cloned.trackFormatInfo = trackFormatInfo.Clone();

      cloned.extractedFilePaths = new Dictionary<Track, string>(this.extractedFilePaths);
      cloned.encodedFilePaths = new Dictionary<Track, string>(this.encodedFilePaths);

      cloned.PropertyChanged = null;
      cloned.PropertyChanging = null;

      return cloned;
    }

    public void ClearTracks()
    {
      if (tracks == null)
        return;

      foreach (var track in tracks) {
        track.ClearDirty();
      }
    }

    public void Save(string file)
    {
      var document = new IniDocument(StringComparer.CurrentCulture);

      document.DefaultSection["Version"] = System.Windows.Forms.Application.ProductVersion;

      var extractionConfigSection = document["ExtractionConfig"];

      extractionConfigSection["Product"]                = product.Title;
      extractionConfigSection["BgmSourcePath"]          = bgmSourcePath;
      extractionConfigSection["OutputDirectory"]        = outputDirectory;
      extractionConfigSection["TrackFormatInfo.ZeroPadding"]              = trackFormatInfo.ZeroPadding.ToString();
      extractionConfigSection["TrackFormatInfo.TitleUnificationOptions"]  = trackFormatInfo.TitleUnificationOptions.ToString();
      extractionConfigSection["TrackFormatString"]      = trackFormatString;
      extractionConfigSection["SaveSourceAfterEncode"]  = saveSourceAfterEncode.ToString();
      extractionConfigSection["EncoderConfig"]          = EncoderConfig.IsNullOrEmpty(encoderConfig) ? string.Empty : encoderConfig.Name;
      extractionConfigSection["PlaylistFormat"]         = (playlistFormat == null) ? string.Empty : playlistFormat.Name;

      foreach (var track in tracks) {
        var trackSection = document[string.Format("Track{0}", track.TrackInfo.TrackNumber)];

        //trackSection["TrackNumber"] = track.TrackInfo.TrackNumber.ToString();
        trackSection["Checked"]     = track.Checked.ToString();

        trackSection["FadingStart"]             = track.Effect.FadingStart.ToString();
        trackSection["FadingLength"]            = track.Effect.FadingLength.ToString();
        trackSection["FadingCurveType"]         = track.Effect.FadingCurveType.ToString();
        trackSection["FadingCurveCoefficient"]  = track.Effect.FadingCurveCoefficient.ToString();
        trackSection["TrailingSilenceLength"]   = track.Effect.TrailingSilenceLength.ToString();
      }

      document.Save(file);
    }

    public static ExtractionConfig LoadFrom(string file)
    {
      try {
        if (PathUtils.AreExtensionEqual(file, THxxBGMFadeoutSettingsFileExtension)) {
          try {
            return LoadFromTHxxBGMFadeoutSettings(file);
          }
          catch (NotSupportedException) {
            throw;
          }
          catch (Exception ex) {
            // ignore, try load as extraction config file
#if DEBUG
            Console.Error.WriteLine(ex);
#endif
          }
        }

        return LoadFromCore(file);
      }
      catch (NotSupportedException) {
        throw;
      }
      catch (Exception ex) {
        throw new System.IO.InvalidDataException("invalid extraction config", ex);
      }
    }

    private static ExtractionConfig LoadFromCore(string file)
    {
      var document = IniDocument.Load(file, StringComparer.CurrentCulture);
      var config = new ExtractionConfig();

      var version = ParseVersion(document.DefaultSection["Version"]);

      var extractionConfigSection = document["ExtractionConfig"];

      config.Product = ThbgmExtractor.Products.Find(delegate(ProductInfo product) {
        return product.Title == extractionConfigSection["Product"];
      });

      if (config.product == null)
        throw new NotSupportedException("unsupported product");

      config.bgmSourcePath   = extractionConfigSection["BgmSourcePath"];
      config.outputDirectory = extractionConfigSection["OutputDirectory"];

      config.trackFormatInfo = new TrackFormatInfo();
      config.trackFormatInfo.ZeroPadding              = extractionConfigSection.Get<bool>("TrackFormatInfo.ZeroPadding", config.trackFormatInfo.ZeroPadding, Convert.ToBoolean);
      config.trackFormatInfo.TitleUnificationOptions  = extractionConfigSection.Get<TitleUnificationOptions>("TrackFormatInfo.TitleUnificationOptions",  config.trackFormatInfo.TitleUnificationOptions, EnumUtils.Parse<TitleUnificationOptions>);
      config.TrackFormatString = extractionConfigSection.Get("TrackFormatString", config.trackFormatString);

      config.saveSourceAfterEncode = extractionConfigSection.Get<bool>("SaveSourceAfterEncode", config.saveSourceAfterEncode, Convert.ToBoolean);

      config.encoderConfig = ThbgmExtractor.EncoderConfigs.Find(delegate(EncoderConfig encoderConfig) {
        return encoderConfig.Name == extractionConfigSection["EncoderConfig"];
      }) ?? EncoderConfig.Empty;

      var playlistFormatName = extractionConfigSection["PlaylistFormat"];

      config.playlistFormat = PlaylistFormat.Formats.Find(delegate(PlaylistFormat format) {
        return format.Name == playlistFormatName;
      }) ?? config.playlistFormat;

      var tracks = new List<Track>();

      foreach (var trackInfo in config.product.Tracks) {
        var trackSection = document.Find(string.Format("Track{0}", trackInfo.TrackNumber));

        Track track;

        if (trackSection == null) {
          track = new Track(trackInfo);
        }
        else {
          var effect = Effect.CreateDefault(trackInfo);

          effect.FadingStart            = TimeSpanUtils.ParsePositive(trackSection["FadingStart"], effect.FadingStart);
          effect.FadingLength           = TimeSpanUtils.ParsePositive(trackSection["FadingLength"], effect.FadingLength);
          effect.TrailingSilenceLength  = TimeSpanUtils.ParsePositive(trackSection["TrailingSilenceLength"], effect.TrailingSilenceLength);

          effect.FadingCurveType = trackSection.Get<FadingCurveType>("FadingCurveType", effect.FadingCurveType, EnumUtils.Parse<FadingCurveType>);

          var fadingCurveCoefficient = trackSection.Get<double?>("FadingCurveCoefficient", null, NullableConvert.ToDouble);

          if (fadingCurveCoefficient.HasValue &&
              (double.IsNaN(fadingCurveCoefficient.Value) || double.IsInfinity(fadingCurveCoefficient.Value)))
            fadingCurveCoefficient = null;

          switch (effect.FadingCurveType) {
            case FadingCurveType.Exponential:
            case FadingCurveType.Logarithmic:
              if (fadingCurveCoefficient == null)
                effect.FadingCurveCoefficient = 10.0;
              else if (fadingCurveCoefficient.Value <= 1.0)
                effect.FadingCurveCoefficient = 1.0;
              else
                effect.FadingCurveCoefficient = fadingCurveCoefficient.Value;
              break;
            case FadingCurveType.Trigonometrical:
              if (fadingCurveCoefficient == null)
                effect.FadingCurveCoefficient = 1.0;
              else if (fadingCurveCoefficient.Value <= 0.0)
                effect.FadingCurveCoefficient = 1.0;
              else
                effect.FadingCurveCoefficient = fadingCurveCoefficient.Value;
              break;
            default:
              effect.FadingCurveCoefficient = 1.0;
              break;
          }

          track = new Track(trackInfo, effect, bool.Parse(trackSection["Checked"]));
        }

        tracks.Add(track);
      }

      config.tracks = tracks;

      if (version.Major == 1 && version.Minor == 4) {
        // version 1.4.x extraction config entries
        if (extractionConfigSection.Entries.ContainsKey("PrependTrackNumber")) {
          if (extractionConfigSection.Get<bool>("PrependTrackNumber", true, Convert.ToBoolean))
            config.trackFormatString = "$trackno. $title";
          else
            config.trackFormatString = "$title";

          config.trackFormatInfo.TitleUnificationOptions = TitleUnificationOptions.NoUnify;
        }

        if (extractionConfigSection.Entries.ContainsKey("UseWaveDashInsteadOfFullWidthTilde") || extractionConfigSection.Entries.ContainsKey("UseTwoSpacesInsteadOfIdeographicSpace")) {
          var useWaveDashInsteadOfFullWidthTilde    = extractionConfigSection.Get<bool>("UseWaveDashInsteadOfFullWidthTilde", false, Convert.ToBoolean);
          var useTwoSpacesInsteadOfIdeographicSpace = extractionConfigSection.Get<bool>("UseTwoSpacesInsteadOfIdeographicSpace", false, Convert.ToBoolean);

          if (useWaveDashInsteadOfFullWidthTilde)
            config.trackFormatInfo.TitleUnificationOptions |= TitleUnificationOptions.WaveDash;
          else
            config.trackFormatInfo.TitleUnificationOptions |= TitleUnificationOptions.FullWidthTilde;

          if (useTwoSpacesInsteadOfIdeographicSpace)
            config.trackFormatInfo.TitleUnificationOptions |= TitleUnificationOptions.DoubleSpace;
          else
            config.trackFormatInfo.TitleUnificationOptions |= TitleUnificationOptions.SingleIdeographicSpace;
        }
      }

      return config;
    }

    private static Version ParseVersion(string version)
    {
      try {
        return new Version(version);
      }
      catch {
        return null;
      }
    }

    private bool OnPropertyChanging(string propertyName)
    {
      var e = new CancellablePropertyChangingEventArgs(propertyName);

      OnPropertyChanging(e);

      return !e.Cancel;
    }

    protected virtual void OnPropertyChanging(PropertyChangingEventArgs e)
    {
      var ev = this.PropertyChanging;

      if (ev != null)
        ev(this, e);
    }

    private void OnPropertyChanged(string propertyName)
    {
      OnPropertyChanged(new PropertyChangedEventArgs(propertyName));
    }

    protected virtual void OnPropertyChanged(PropertyChangedEventArgs e)
    {
      var ev = this.PropertyChanged;

      if (ev != null)
        ev(this, e);
    }

#region "loading/saving THxxBGMFadeoutSettings"
    private static ExtractionConfig LoadFromTHxxBGMFadeoutSettings(string fsxFile)
    {
      var xml = new XmlDocument();

      xml.Load(fsxFile);

      var config = new ExtractionConfig();

      var productGuid = xml.GetSingleNodeValueOf<Guid?>("/THxxBGMFadeoutSettings/Guid/text()", NullableConvert.ToGuidIgnoreException);

      if (productGuid == null)
        config.product = null;
      else
        config.product = ThbgmExtractor.Products.Find(delegate(ProductInfo product) {
          return product.THxxBGMProductGuid == productGuid.Value;
        });

      if (config.product == null)
        throw new NotSupportedException("unsupported product");

      config.bgmSourcePath = config.product.InstalledOrDefaultBgmSourcePath;

      // translate '/THxxBGMFadeoutSettings/Data/Item'
      var tracks = new List<Track>();

      for (var i = 0; i < config.product.Tracks.Count(); i++) {
        var trackInfo = config.product.Tracks[i];
        var itemNode = xml.SelectSingleNode(string.Format("/THxxBGMFadeoutSettings/Data/Item[{0}]", i + 1));
        Track track;

        if (itemNode == null) {
          track = new Track(trackInfo);
        }
        else {
          var timesToRepeat = itemNode.GetSingleNodeValueOf<int?>("LoopCount/text()", NullableConvert.ToInt32IgnoreException) ?? Effect.DefaultTimesToRepeat;

          // translate FadeoutTime and BgmEnd
          var fadingLength = trackInfo.StreamFormat.ToTimeSpan(itemNode.GetSingleNodeValueOf<int?>("FadeoutTime/text()", NullableConvert.ToInt32IgnoreException) ?? 0);
          var fadingOffset = trackInfo.StreamFormat.ToTimeSpan(itemNode.GetSingleNodeValueOf<int?>("BgmEnd/text()", NullableConvert.ToInt32IgnoreException) ?? 0);

          var effect = new Effect();

          effect.SetLength(trackInfo, timesToRepeat, fadingLength, -fadingLength + fadingOffset, TimeSpan.Zero);

          // translate FadeoutPattern
          switch (itemNode.GetSingleNodeValueOf<int?>("FadeoutPattern/text()", NullableConvert.ToInt32IgnoreException) ?? 1) {
            case 2: // linear
              effect.FadingCurveType = FadingCurveType.Linear;
              break;

            case 3:
              // (linear + exponential) / 2
              effect.FadingCurveType = FadingCurveType.Exponential;
              effect.FadingCurveCoefficient = 10.0; // XXX: approx
              break;

            case 4:
              effect.FadingCurveType = FadingCurveType.Trigonometrical;
              effect.FadingCurveCoefficient = 1.0;
              break;

            case 1: // exponential
            default:
              effect.FadingCurveType = FadingCurveType.Exponential;
              effect.FadingCurveCoefficient = 100.0; // XXX: approx
              break;
          }

          // TODO: translate BgmBegin and BgmEnd

          track = new Track(trackInfo, effect, true);
        }

        tracks.Add(track);
      }

      config.tracks = tracks;

      return config;
    }

    public void SaveAsTHxxBGMFadeoutSettings(string file)
    {
      if (Guid.Empty.Equals(product.THxxBGMProductGuid))
        throw new InvalidOperationException(string.Format("THxxBMG does not support product '{0}'", product.Title));

      var xml = new XmlDocument();

      xml.AppendChild(xml.CreateXmlDeclaration("1.0", "UTF-8", "no")); // XXX: why standalone='no'?

      // /THxxBGMFadeoutSettings
      var root = xml.AppendChild(xml.CreateElement("THxxBGMFadeoutSettings"));

      // /THxxBGMFadeoutSettings/GameTitle
      root.AppendChild(xml.CreateElement("GameTitle")).AppendChild(xml.CreateTextNode(product.ShortName));

      // /THxxBGMFadeoutSettings/Guid (GUID with braces)
      root.AppendChild(xml.CreateElement("Guid")).AppendChild(xml.CreateTextNode(product.THxxBGMProductGuid.ToString("B").ToUpperInvariant()));

      // /THxxBGMFadeoutSettings/Data
      var data = root.AppendChild(xml.CreateElement("Data"));

      // /THxxBGMFadeoutSettings/Data/Item
      foreach (var track in tracks) {
        int fadeoutPattern;

        switch (track.Effect.FadingCurveType) {
          case FadingCurveType.Linear: fadeoutPattern = 2; break;
          case FadingCurveType.Trigonometrical: fadeoutPattern = 4; break;

          case FadingCurveType.Exponential:
          default:
            // THxxBGM default
            fadeoutPattern = 1;
            break;
        }

        var loopCount = track.TrackInfo.CalculateRepeatedTimes(track.Effect.TotalLength);
        var fadeoutTime = track.TrackInfo.StreamFormat.ToAlignedByteCount(track.Effect.FadingLength);
        var repeatedTimes = track.TrackInfo.CalculateRepeatedTimes(track.Effect.TotalLength);

        // THxxBGM does not support LoopCount=0?
        if (loopCount == 0) {
          loopCount += 1;
          repeatedTimes += 1;
        }

        var fadingOffset = track.Effect.TotalLength - track.TrackInfo.CalculateRepeatPosition(repeatedTimes);
        var bgmEnd = track.TrackInfo.StreamFormat.ToAlignedByteCount(fadingOffset);

        var item = data.AppendChild(xml.CreateElement("Item"));

        item.AppendChild(xml.CreateElement("LoopCount")).AppendChild(xml.CreateTextNode(XmlConvert.ToString(loopCount)));
        item.AppendChild(xml.CreateElement("FadeoutPattern")).AppendChild(xml.CreateTextNode(XmlConvert.ToString(fadeoutPattern)));
        item.AppendChild(xml.CreateElement("BgmBegin")).AppendChild(xml.CreateTextNode("0")); // ThbgmExtractor unsupported feature
        item.AppendChild(xml.CreateElement("BgmEnd")).AppendChild(xml.CreateTextNode(XmlConvert.ToString(bgmEnd)));
        item.AppendChild(xml.CreateElement("FadeoutTime")).AppendChild(xml.CreateTextNode(XmlConvert.ToString(fadeoutTime)));
      }

      using (var stream = File.OpenWrite(file)) {
        var settings = new XmlWriterSettings();

        settings.Encoding = new UTF8Encoding(false); // output no preamble
        settings.Indent = true;
        settings.IndentChars = "\t";
        settings.NewLineChars = Environment.NewLine;

        var writer = XmlWriter.Create(stream, settings);

        xml.WriteTo(writer);

        writer.Flush();
      }
    }
#endregion
  }
}
