// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Windows.Forms;

using Smdn.Collections;
using Smdn.IO;
using Smdn.Formats.Playlists;
using Smdn.Windows.Forms.ThbgmPlayer;

namespace Smdn.Applications.ThbgmExtractor {
  internal class ExtractionPreconfigurator {
    public ExtractionConfig Config {
      get { return config; }
    }

    public ExtractionPreconfigurator(MainDialog owner, ExtractionConfig baseConfig)
    {
      if (owner == null)
        throw new ArgumentNullException("owner");

      this.mainDialog = owner;
      this.config = baseConfig.Clone();
    }

    private bool CheckAndAlertBmgSourceExistence(string bgmSourcePath)
    {
      if (File.Exists(bgmSourcePath))
        return true;

      var message = string.IsNullOrEmpty(bgmSourcePath)
        ? ThbgmExtractor.GetText("MessageBox_CheckAndAlertBmgSourceExistence_BgmSourceNotSet")
        : ThbgmExtractor.GetText("MessageBox_CheckAndAlertBmgSourceExistence_BgmSourceNotFound", bgmSourcePath);

      MessageBox.Show(mainDialog, message, Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Exclamation);

      mainDialog.FocusBgmSourcePath();

      return false;
    }

    private string SelectOutputFile(string initialOutputFile, ref EncoderConfig encoderConfig)
    {
      var encoders = new List<EncoderConfig>(ThbgmExtractor.EncoderConfigs);

      using (var dialog = new SaveFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(encoders.ConvertAll(delegate(EncoderConfig config) {
          return new FileDialogFilter.Filter(EncoderConfig.IsNullOrEmpty(config)
                                               ? ThbgmExtractor.GetText("EncoderConfig_NonEncoded")
                                               : config.Name,
                                             "*" + config.OutputExtension);
        }));
        dialog.FilterIndex = encoders.IndexOf(encoderConfig) + 1; // 1-based
        dialog.FileName = initialOutputFile;
        dialog.InitialDirectory = Path.GetDirectoryName(initialOutputFile);
        dialog.RestoreDirectory = true;
        dialog.DefaultExt = encoderConfig.OutputExtension.Substring(1); // remove period
        dialog.AddExtension = true;

        if (dialog.ShowDialog(mainDialog) == DialogResult.OK) {
          encoderConfig = encoders[dialog.FilterIndex - 1]; // 1-based
          return dialog.FileName;
        }
        else {
          encoderConfig = EncoderConfig.Empty;
          return null; // as canceled
        }
      }
    }

    private string SelectOutputDirectory(string initialOutputDirectory)
    {
      using (var dialog = new FolderBrowserDialog()) {
        dialog.Description = ThbgmExtractor.GetText("FolderBrowserDialog_SelectOutputDirectory_Description");
        dialog.RootFolder = Environment.SpecialFolder.Desktop;
        dialog.SelectedPath = initialOutputDirectory;
        dialog.ShowNewFolderButton = true;

        if (dialog.ShowDialog(mainDialog) == DialogResult.OK)
          return dialog.SelectedPath;
        else
          return null; // as canceled
      }
    }

    private bool CheckAndConfirmCreateDirectory(string directory)
    {
      if (Directory.Exists(directory))
        return true;

      var result = MessageBox.Show(mainDialog,
                                   ThbgmExtractor.GetText("MessageBox_CheckAndConfirmCreateDirectory", directory),
                                   Application.ProductName,
                                   MessageBoxButtons.YesNo,
                                   MessageBoxIcon.Question,
                                   MessageBoxDefaultButton.Button1);

      if (result != DialogResult.Yes)
        return false;

      Directory.CreateDirectory(directory);

      return true;
    }

    private string SelectPlaylistPath(string initialOutputFilePath, PlaylistFormat format)
    {
      using (var dialog = new SaveFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
          new[] {format.Name, "*" + format.DefaultExtension},
          new[] {Texts.GetText("AllFiles"), "*.*"},
        });

        dialog.FilterIndex = 0;
        dialog.FileName = initialOutputFilePath;
        dialog.InitialDirectory = Path.GetDirectoryName(initialOutputFilePath);
        dialog.RestoreDirectory = true;

        if (dialog.ShowDialog(mainDialog) == DialogResult.OK)
          return dialog.FileName;
        else
          return null; // as canceled
      }
    }

    private bool CheckAndConfirmOverwriteFiles(IEnumerable<string> files)
    {
      var fileExists = false;

      foreach (var file in files) {
        if (File.Exists(file)) {
          fileExists = true;
          break;
        }
      }

      if (!fileExists)
        return true;

      var result = MessageBox.Show(mainDialog,
                                   ThbgmExtractor.GetText("MessageBox_CheckAndConfirmOverwriteFiles"),
                                   Application.ProductName,
                                   MessageBoxButtons.YesNo,
                                   MessageBoxIcon.Question,
                                   MessageBoxDefaultButton.Button2);

      if (result == DialogResult.Yes)
        return true; // overwrite if exists
      else
        return false;
    }

    public bool Configure(bool singleTrack)
    {
      /*
       * bmg source path
       */
      if (!CheckAndAlertBmgSourceExistence(config.BgmSourcePath))
        return false;

      if (singleTrack)
        return ConfigureSingleTrack();
      else
        return Configure();
    }

    private bool ConfigureSingleTrack()
    {
      var track = config.Tracks.First();

      config.PlaylistFormat = null;

      /*
       * output file
       */
      var encoderConfig = config.EncoderConfig ?? EncoderConfig.Empty;
      var outputDirectory = config.OutputDirectory;

      if (string.IsNullOrEmpty(outputDirectory))
        outputDirectory = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);

      var outputFile = Path.Combine(outputDirectory,
                                    PathUtils.ReplaceInvalidFileNameCharsWithBlanks(track.TrackInfo.ToString(config.TrackFormatString, config.TrackFormatInfo)));

      outputFile = SelectOutputFile(outputFile, ref encoderConfig);

      if (outputFile == null)
        return false; // canceled

      config.OutputDirectory = Path.GetDirectoryName(outputFile);
      config.EncoderConfig = encoderConfig;

      /*
       * set tracks and paths
       */
      config.Tracks = new Track[] {track};

      config.ExtractedFilePaths.Clear();
      config.ExtractedFilePaths.Add(track, Path.ChangeExtension(outputFile, ".wav"));

      config.EncodedFilePaths.Clear();
      config.EncodedFilePaths.Add(track, outputFile);

      /*
       * done
       */
      return true;
    }

    private bool Configure()
    {
      /*
       * output directory
       */
      var outputDirectory = config.OutputDirectory;

      if (string.IsNullOrEmpty(outputDirectory)) {
        var dirname = config.Product.GetUnifiedTitle(config.TrackFormatInfo.TitleUnificationOptions);

        outputDirectory = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.Desktop),
                                       PathUtils.ReplaceInvalidPathCharsWithBlanks(dirname));
      }

      outputDirectory = SelectOutputDirectory(outputDirectory);

      if (outputDirectory == null)
        return false; // canceled

      config.OutputDirectory = outputDirectory;

      /*
       * create output directory
       */
      if (!CheckAndConfirmCreateDirectory(config.OutputDirectory))
        return false; // canceled

      /*
       * playlist path
       */
      string playlistPath = null;

      if (config.PlaylistFormat != null) {
        playlistPath = Path.Combine(config.OutputDirectory,
                                    config.Product.GetUnifiedTitle(config.TrackFormatInfo.TitleUnificationOptions) + config.PlaylistFormat.DefaultExtension);

        playlistPath = SelectPlaylistPath(playlistPath, config.PlaylistFormat);

        if (playlistPath == null)
          return false; // canceled
      }

      /*
       * extracted/encoded tracks and paths
       */
      var extractTracks = new List<Track>();
      var extractedFilePaths = new Dictionary<Track, string>();
      var encodedFilePaths = new Dictionary<Track, string>();
      var doEncode = !EncoderConfig.IsNullOrEmpty(config.EncoderConfig);

      foreach (var track in config.Tracks) {
        if (!track.Checked)
          continue;

        var t = track.Clone();

        extractTracks.Add(t);

        var outputFileName = PathUtils.ReplaceInvalidFileNameCharsWithBlanks(t.TrackInfo.ToString(config.TrackFormatString, config.TrackFormatInfo));

        extractedFilePaths.Add(t, Path.Combine(outputDirectory, outputFileName + ".wav"));

        if (doEncode)
          encodedFilePaths.Add(t, Path.Combine(outputDirectory, outputFileName + config.EncoderConfig.OutputExtension));
      }

      /*
       * check overwriting
       */
      var allOutputFiles = new List<string>();

      allOutputFiles.AddRange(extractedFilePaths.Values);
      allOutputFiles.AddRange(encodedFilePaths.Values);

      if (!CheckAndConfirmOverwriteFiles(allOutputFiles))
        return false; // canceled

      /*
       * create and save playlist
       */
      if (playlistPath != null) {
        try {
          TrackPlaylist.Save(playlistPath, doEncode ? encodedFilePaths : extractedFilePaths, config.PlaylistFormat);
        }
        catch (Exception ex) {
          MessageBox.Show(mainDialog,
                          string.Format("save playlist failed: {0}", ex),
                          Application.ProductName,
                          MessageBoxButtons.OK,
                          MessageBoxIcon.Error);
        }
      }

      /*
       * set tracks and paths
       */
      config.Tracks = extractTracks;

      config.ExtractedFilePaths.Clear();

      foreach (var pair in extractedFilePaths) {
        config.ExtractedFilePaths.Add(pair.Key, pair.Value);
      }

      config.EncodedFilePaths.Clear();

      foreach (var pair in encodedFilePaths) {
        config.EncodedFilePaths.Add(pair.Key, pair.Value);
      }

      /*
       * done
       */
      return true;
    }

    private MainDialog mainDialog;
    private ExtractionConfig config;
  }
}
