using System;
using System.Collections.Generic;
using System.Windows.Forms;

using Smdn.Formats.Thbgm;

namespace Smdn.Applications.ThbgmExtractor {
  public partial class FileNameEditDialog : Form {
    private class TrackInfoFormatToolStripMenuItem : ToolStripMenuItem {
      public string Format {
        get; private set;
      }

      public TrackInfoFormatToolStripMenuItem(string format, string text, EventHandler onClick)
        : base(text, null, onClick)
      {
        this.Format = format;
      }
    }

    private class TrackInfoFormatPresetToolStripMenuItem : TrackInfoFormatToolStripMenuItem {
      public TrackInfoFormatPresetToolStripMenuItem(string format, string text, EventHandler onClick)
        : base(format, text, onClick)
      {
      }
    }

    public string Format {
      get { return textBoxFormat.Text; }
      set { textBoxFormat.Text = value; }
    }

    public TrackFormatInfo TrackFormatInfo {
      get { return trackFormatInfo; }
      set
      {
        if (trackFormatInfo == value)
          return;

        trackFormatInfo = value ?? new TrackFormatInfo();

        if (lockChanged)
          return;

        try {
          lockChanged = true;

          OnTrackFormatInfoChanged();
        }
        finally {
          lockChanged = false;
        }
      }
    }

    public TrackInfo TrackForExample {
      get; set;
    }

    public FileNameEditDialog()
    {
      InitializeComponent();
      Initialize();
    }

    private void Initialize()
    {
      var presetsMenuItem = new ToolStripMenuItem(ThbgmExtractor.GetText("FileNameEditDialog_Presets"));

      foreach (var pair in new[] {
        new {Format = "$title",                       Text = ThbgmExtractor.GetText("FileNameFormat_Preset_TitleOnly")},
        new {Format = "$trackno. $title",             Text = ThbgmExtractor.GetText("FileNameFormat_Preset_MusicRoomStyle1")},
        new {Format = "No$trackno. $title",           Text = ThbgmExtractor.GetText("FileNameFormat_Preset_MusicRoomStyle2")},
        new {Format = "$prefix_$trackno",             Text = ThbgmExtractor.GetText("FileNameFormat_Preset_Th06BgmFileStyle")},
        new {Format = "$pshortest$trackno：$title",   Text = ThbgmExtractor.GetText("FileNameFormat_Preset_THxxBGMStyle")},
      }) {
        presetsMenuItem.DropDownItems.Add(new TrackInfoFormatPresetToolStripMenuItem(pair.Format, pair.Text, contextMenuStripFormats_ItemsClick));
      }

      contextMenuStripFormats.Items.AddRange(new ToolStripItem[] {
        presetsMenuItem,
        new ToolStripSeparator(),
      });

      foreach (var pair in new[] {
        new {Format = "$title",     Text = ThbgmExtractor.GetText("FileNameFormat_title")},
        new {Format = "$trackno",   Text = ThbgmExtractor.GetText("FileNameFormat_trackno")},
        new {Format = "$tracks",    Text = ThbgmExtractor.GetText("FileNameFormat_tracks")},
        new {Format = "$creator",   Text = ThbgmExtractor.GetText("FileNameFormat_creator")},
        new {Format = "$product",   Text = ThbgmExtractor.GetText("FileNameFormat_product")},
        new {Format = "$pshort1",   Text = ThbgmExtractor.GetText("FileNameFormat_pshort1")},
        new {Format = "$pshort2",   Text = ThbgmExtractor.GetText("FileNameFormat_pshort2")},
        new {Format = "$pshortest", Text = ThbgmExtractor.GetText("FileNameFormat_pshortest")},
        new {Format = "$prefix",    Text = ThbgmExtractor.GetText("FileNameFormat_prefix")},
        new {Format = "$ryear",     Text = ThbgmExtractor.GetText("FileNameFormat_ryear")},
        new {Format = "$rmonth",    Text = ThbgmExtractor.GetText("FileNameFormat_rmonth")},
        new {Format = "$rday",      Text = ThbgmExtractor.GetText("FileNameFormat_rday")},
      }) {
        contextMenuStripFormats.Items.Add(new TrackInfoFormatToolStripMenuItem(pair.Format, pair.Text, contextMenuStripFormats_ItemsClick));
      }

      contextMenuStripFormats.Items.AddRange(new ToolStripItem[] {
        new ToolStripSeparator(),
        new ToolStripMenuItem(ThbgmExtractor.GetText("FileNameEditDialog_Clear"), null, delegate {
          textBoxFormat.Clear();
        }),
      });
    }

    protected override void OnShown(EventArgs e)
    {
      OnTrackFormatInfoChanged();

      base.OnShown(e);
    }

    private void buttonSetFormat_Click(object sender, EventArgs e)
    {
      var control = sender as Control;

      contextMenuStripFormats.Show(control, control.PointToClient(Cursor.Position));
    }

    private void textBoxFormat_TextChanged(object sender, EventArgs e)
    {
      var formattedFilename = (TrackForExample == null)
        ? textBoxFormat.Text
        : TrackForExample.ToString(textBoxFormat.Text, trackFormatInfo);

      labelExample.Text = ThbgmExtractor.GetText("FileNameEditDialog_ExampleFormatString", formattedFilename);
    }

    private void contextMenuStripFormats_ItemsClick(object sender, EventArgs e)
    {
      if (sender is TrackInfoFormatPresetToolStripMenuItem) {
        var item = sender as TrackInfoFormatPresetToolStripMenuItem;

        textBoxFormat.Text = item.Format;
      }
      else if (sender is TrackInfoFormatToolStripMenuItem) {
        var item = sender as TrackInfoFormatToolStripMenuItem;

        if (0 < textBoxFormat.TextLength)
          textBoxFormat.Text = string.Format("{0} {1}", textBoxFormat.Text, item.Format);
        else
          textBoxFormat.Text = item.Format;
      }
    }

    private bool lockChanged = false;

    private void checkBoxZeroPadding_CheckedChanged(object sender, EventArgs e)
    {
      if (lockChanged)
        return;

      try {
        lockChanged = true;

        trackFormatInfo.ZeroPadding = checkBoxZeroPadding.Checked;

        OnTrackFormatInfoChanged();
      }
      finally {
        lockChanged = false;
      }
    }

    private void checkBoxUseWaveDashInsteadOfFullWidthTilde_CheckedChanged(object sender, EventArgs e)
    {
      if (lockChanged)
        return;

      try {
        lockChanged = true;

        if (checkBoxUseWaveDashInsteadOfFullWidthTilde.Checked)
          trackFormatInfo.TitleUnificationOptions = (trackFormatInfo.TitleUnificationOptions & ~TitleUnificationOptions.TitleDelimiterOptionMask) | TitleUnificationOptions.WaveDash;
        else
          trackFormatInfo.TitleUnificationOptions = (trackFormatInfo.TitleUnificationOptions & ~TitleUnificationOptions.TitleDelimiterOptionMask) | TitleUnificationOptions.FullWidthTilde;

        OnTrackFormatInfoChanged();
      }
      finally {
        lockChanged = false;
      }
    }

    private void checkBoxUseTwoSpacesInsteadOfIdeographicSpace_CheckedChanged(object sender, EventArgs e)
    {
      if (lockChanged)
        return;

      try {
        lockChanged = true;

        if (checkBoxUseTwoSpacesInsteadOfIdeographicSpace.Checked)
          trackFormatInfo.TitleUnificationOptions = (trackFormatInfo.TitleUnificationOptions & ~TitleUnificationOptions.SpaceOptionMask) | TitleUnificationOptions.DoubleSpace;
        else
          trackFormatInfo.TitleUnificationOptions = (trackFormatInfo.TitleUnificationOptions & ~TitleUnificationOptions.SpaceOptionMask) | TitleUnificationOptions.SingleIdeographicSpace;

        OnTrackFormatInfoChanged();
      }
      finally {
        lockChanged = false;
      }
    }

    private void OnTrackFormatInfoChanged()
    {
      checkBoxZeroPadding.Checked = trackFormatInfo.ZeroPadding;

      checkBoxUseWaveDashInsteadOfFullWidthTilde.Checked = ((trackFormatInfo.TitleUnificationOptions & TitleUnificationOptions.TitleDelimiterOptionMask) == TitleUnificationOptions.WaveDash);
      checkBoxUseTwoSpacesInsteadOfIdeographicSpace.Checked = ((trackFormatInfo.TitleUnificationOptions & TitleUnificationOptions.SpaceOptionMask) == TitleUnificationOptions.DoubleSpace);

      textBoxFormat_TextChanged(null, EventArgs.Empty);
    }

    private void buttonOK_Click(object sender, EventArgs e)
    {
      DialogResult = DialogResult.OK;

      Close();
    }

    private TrackFormatInfo trackFormatInfo = new TrackFormatInfo();
  }
}