using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Threading;
using System.Windows.Forms;

using Smdn.Collections;
using Smdn.Formats.Playlists;
using Smdn.Formats.Thbgm;
using Smdn.IO;
using Smdn.Windows.Forms.ThbgmPlayer;

namespace Smdn.Applications.ThbgmExtractor {
  public partial class MainDialog : Form {
    private class EncoderConfigToolStripMenuItem : ToolStripMenuItem {
      public EncoderConfig EncoderConfig {
        get; private set;
      }

      public EncoderConfigToolStripMenuItem(EncoderConfig encoderConfig, EventHandler onClick)
        : base(encoderConfig.Name, null, onClick)
      {
        this.EncoderConfig = encoderConfig;
      }
    }

    private class PlaylistFormatToolStripMenuItem : ToolStripMenuItem {
      public PlaylistFormat PlaylistFormat {
        get; private set;
      }

      public PlaylistFormatToolStripMenuItem(PlaylistFormat playlistFormat, EventHandler onClick)
        : base(playlistFormat == null
                 ? ThbgmExtractor.GetText("PlaylistFormat_Null_Name")
                 : string.Format("{0} ({1})", playlistFormat.Name, playlistFormat.DefaultExtension),
               null, onClick)
      {
        this.PlaylistFormat = playlistFormat;
      }
    }

    private class FadingCuverTypeStripMenuItem : ToolStripMenuItem {
      public FadingCurveType FadingCurveType {
        get; private set;
      }

      public FadingCuverTypeStripMenuItem(FadingCurveType fadingCurveType, EventHandler onClick)
        : base(Texts.GetText(fadingCurveType), null, onClick)
      {
        this.FadingCurveType = fadingCurveType;
      }
    }

    private class TimesToRepeatStripMenuItem : ToolStripMenuItem {
      public int TimesToRepeat {
        get; private set;
      }

      public TimesToRepeatStripMenuItem(int timesToRepeat, EventHandler onClick)
        : base(ThbgmExtractor.GetText("TimesToRepeatStripMenuItem_FormatString", timesToRepeat), null, onClick)
      {
        this.TimesToRepeat = timesToRepeat;
      }
    }

    private class ExtractionConfigFileStripMenuItem : ToolStripMenuItem {
      public string File {
        get; private set;
      }

      public ExtractionConfigFileStripMenuItem(string file, EventHandler onClick)
        : base(Path.GetFileName(file), null, onClick)
      {
        
        this.File = file;
      }
    }

    public string CurrentConfigFilePath {
      get { return currentConfigFilePath; }
      set
      {
        if (currentConfigFilePath == value)
          return;

        currentConfigFilePath = value;

        UpdateText();
      }
    }

    public bool IsEdittingExtractionConfigDirty {
      get { return isEdittingExtractionConfigDirty; }
      set
      {
        isEdittingExtractionConfigDirty = value;

        UpdateText();
      }
    }

    public MainDialog()
    {
      InitializeComponent();
      Initialize();

      if (Runtime.IsRunningOnMono) {
        // XXX
        this.ClientSize = new Size(784, 563);
        productSelector.Width = this.ClientSize.Width;
        splitContainerMain.Size = new Size(784, 486);
        splitContainerMain.SplitterDistance = 300;
        splitContainerExtraction.Size = new Size(784, 241);
        splitContainerExtraction.SplitterDistance = 490;
      }

      UpdateText();
    }

    private void Initialize()
    {
      textBoxLog.MaxLength = int.MaxValue;

      timerLog.Start();

      splitContainerExtraction.Panel2Collapsed = true;

      productSelector.SelectedProductChanged += delegate {
        edittingExtractionConfig.Product = productSelector.SelectedProduct;

        trackListView.Product = edittingExtractionConfig.Product;
      };

      productSelector.BgmSourcePathChanged += delegate {
        edittingExtractionConfig.BgmSourcePath = productSelector.BgmSourcePath;

        player.BgmSourcePath = edittingExtractionConfig.BgmSourcePath;
      };

      var sortedByReleaseDateProducts = new List<ProductInfo>(ThbgmExtractor.Products);

      sortedByReleaseDateProducts.Sort(delegate(ProductInfo x, ProductInfo y) {
        return -1/*negate, latest first*/ * DateTime.Compare(x.ReleaseDate, y.ReleaseDate);
      });

      productSelector.Products = sortedByReleaseDateProducts;

      var encoderConfigInsertIndex = toolStripMenuItemExtract_EncoderConfigs.DropDownItems.IndexOf(toolStripMenuItemExtract_EncoderConfigs_Separator);
      var encoderConfigOnClick = new EventHandler(delegate(object sender, EventArgs e) {
        edittingExtractionConfig.EncoderConfig = (sender as EncoderConfigToolStripMenuItem).EncoderConfig;
        IsEdittingExtractionConfigDirty = true;

        SetSelectedEncoderConfigCheckedState();
      });

      foreach (var encoderConfig in ThbgmExtractor.EncoderConfigs) {
        toolStripMenuItemExtract_EncoderConfigs.DropDownItems.Insert(encoderConfigInsertIndex, new EncoderConfigToolStripMenuItem(encoderConfig, encoderConfigOnClick));

        encoderConfigInsertIndex++;
      }

      var playlistFormatOnClick = new EventHandler(delegate(object sender, EventArgs e) {
        Config.ExtractionConfigDefaults.PlaylistFormat =
        edittingExtractionConfig.PlaylistFormat = (sender as PlaylistFormatToolStripMenuItem).PlaylistFormat;
        IsEdittingExtractionConfigDirty = true;

        SetSelectedPlaylistFormatCheckedState();
      });

      toolStripMenuItemExtract_PlaylistFormats.DropDownItems.Add(new PlaylistFormatToolStripMenuItem(null, playlistFormatOnClick));
      toolStripMenuItemExtract_PlaylistFormats.DropDownItems.Add(new ToolStripSeparator());

      foreach (var playlistFormat in PlaylistFormat.Formats) {
        if (playlistFormat.Format != null)
          toolStripMenuItemExtract_PlaylistFormats.DropDownItems.Add(new PlaylistFormatToolStripMenuItem(playlistFormat, playlistFormatOnClick));
      }

      toolStripMenuItemExtract.DropDownOpening += delegate {
        toolStripMenuItemExtract_Extract.Enabled = !backgroundWorkerExtraction.IsBusy;
        toolStripMenuItemExtract_ExtractTrack.Enabled = (trackListView.SelectedTrack != null) && !backgroundWorkerExtraction.IsBusy;
      };

      toolStripMenuItemExtract_ExtractTrack.Enabled =
      toolStripMenuItemTrackList_ExtractTrack.Enabled = false;

      trackListView.ProductChanged += delegate {
        edittingExtractionConfig.Tracks = trackListView.Tracks;
        toolStripMenuItemTrackList_ExtractTrack.Enabled =
        toolStripMenuItemExtract_ExtractTrack.Enabled = false;
        player.Track = null;
        UpdateText();
      };
      trackListView.SelectedIndexChanged += delegate {
        toolStripMenuItemTrackList_ExtractTrack.Enabled =
        toolStripMenuItemExtract_ExtractTrack.Enabled = (trackListView.SelectedTrack != null) && !backgroundWorkerExtraction.IsBusy;
        player.Track = trackListView.SelectedTrack;
      };
      trackListView.TrackEffectPropertyChanged += delegate {
        UpdateText();
      };

      trackListView.Product = null;
      trackListView.Product = productSelector.SelectedProduct;

      toolStripMenuItemExtract_EncoderConfigs_SaveSourceAfterEncode.Click += delegate {
        Config.ExtractionConfigDefaults.SaveSourceAfterEncode =
        edittingExtractionConfig.SaveSourceAfterEncode = toolStripMenuItemExtract_EncoderConfigs_SaveSourceAfterEncode.Checked;

        IsEdittingExtractionConfigDirty = true;
      };

      foreach (var fadingCurveType in new[] {
        FadingCurveType.Linear,
        FadingCurveType.Exponential,
        FadingCurveType.Logarithmic,
        FadingCurveType.Trigonometrical,
        FadingCurveType.None,
      }) {
        toolStripMenuItemEdit_FadingCurveType.DropDownItems.Add(new FadingCuverTypeStripMenuItem(fadingCurveType, delegate(object sender, EventArgs e) {
          if (trackListView.SelectedTrack != null)
            trackListView.SelectedTrack.Effect.FadingCurveType = (sender as FadingCuverTypeStripMenuItem).FadingCurveType;
        }));
      }

      toolStripMenuItemEdit_FadingCurveType.DropDownOpening += delegate {
        var selectedFadingCurveType = (trackListView.SelectedTrack == null)
          ? (FadingCurveType?)null
          : (FadingCurveType?)trackListView.SelectedTrack.Effect.FadingCurveType;

        foreach (FadingCuverTypeStripMenuItem item in toolStripMenuItemEdit_FadingCurveType.DropDownItems) {
          item.Checked = (item.FadingCurveType == selectedFadingCurveType);
        }
      };

      for (var i = 0; i < 10; i++) {
        toolStripMenuItemEdit_TimesToRepeat.DropDownItems.Add(new TimesToRepeatStripMenuItem(i, delegate(object sender, EventArgs e) {
          if (trackListView.SelectedTrack == null)
            return;

          var track = trackListView.SelectedTrack;
          var fadingOffset = track.Effect.FadingStart - track.TrackInfo.CalculateRepeatPosition(track.TrackInfo.CalculateRepeatedTimes(track.Effect.TotalLength));

          track.Effect.SetLength(track.TrackInfo, (sender as TimesToRepeatStripMenuItem).TimesToRepeat, track.Effect.FadingLength, fadingOffset, track.Effect.TrailingSilenceLength);
        }));
      }

      toolStripMenuItemEdit_TimesToRepeat.DropDownOpening += delegate {
        var selectedTimesToRepeat = (trackListView.SelectedTrack == null)
          ? -1
          : trackListView.SelectedTrack.TrackInfo.CalculateRepeatedTimes(trackListView.SelectedTrack.Effect.TotalLength);

        foreach (TimesToRepeatStripMenuItem item in toolStripMenuItemEdit_TimesToRepeat.DropDownItems) {
          item.Checked = (item.TimesToRepeat == selectedTimesToRepeat);
        }
      };

      toolStripMenuItemEdit_SelectAll.Click += delegate {
        trackListView.CheckAll();
      };

      toolStripMenuItemEdit_UnselectAll.Click += delegate {
        trackListView.UncheckAll();
      };

      foreach (ToolStripItem item in toolStripMenuItemEdit.DropDownItems) {
        editToolStripItems.Add(item);
      }

      foreach (ToolStripItem item in contextMenuStripTrackList.Items) {
        trackListToolStripItems.Add(item);
      }

      contextMenuStripTrackList.Items.Add(contextMenuDummyItem);

      contextMenuStripTrackList.Opening += delegate {
        var enabled = (trackListView.SelectedTrack != null);

        toolStripMenuItemEdit_FadingCurveType.Enabled = enabled;
        toolStripMenuItemEdit_TimesToRepeat.Enabled = enabled;

        toolStripMenuItemEdit.DropDownItems.Clear();
        contextMenuStripTrackList.Items.Clear();
        contextMenuStripTrackList.Items.AddRange(trackListToolStripItems.ToArray());
        contextMenuStripTrackList.Items.AddRange(editToolStripItems.ToArray());
      };

      contextMenuStripTrackList.Closed += delegate {
        contextMenuStripTrackList.Items.Clear();
        contextMenuStripTrackList.Items.Add(contextMenuDummyItem);
        toolStripMenuItemEdit.DropDownItems.Clear();
        toolStripMenuItemEdit.DropDownItems.AddRange(editToolStripItems.ToArray());
      };

      toolStripMenuItemEdit.DropDownOpening += delegate {
        var enabled = (trackListView.SelectedTrack != null);

        toolStripMenuItemEdit_FadingCurveType.Enabled = enabled;
        toolStripMenuItemEdit_TimesToRepeat.Enabled = enabled;

        contextMenuStripTrackList.Items.Clear();
        contextMenuStripTrackList.Items.Add(contextMenuDummyItem);
        toolStripMenuItemEdit.DropDownItems.Clear();
        toolStripMenuItemEdit.DropDownItems.AddRange(editToolStripItems.ToArray());
      };

      toolStripMenuItemEdit.DropDownClosed += delegate {
        toolStripMenuItemEdit.DropDownItems.Clear();
        contextMenuStripTrackList.Items.Clear();
        contextMenuStripTrackList.Items.AddRange(trackListToolStripItems.ToArray());
        contextMenuStripTrackList.Items.AddRange(editToolStripItems.ToArray());
      };

      // drop down dummy item
      toolStripMenuItemFile_Export_TitleFiles.DropDownItems.Add(new ToolStripSeparator());
      toolStripMenuItemFile_RecentFiles.DropDownItems.Add(new ToolStripSeparator());

      var initialExtractionConfig = new ExtractionConfig();

      initialExtractionConfig.Product = sortedByReleaseDateProducts[0];
      initialExtractionConfig.BgmSourcePath = sortedByReleaseDateProducts[0].InstalledOrDefaultBgmSourcePath;
      initialExtractionConfig.OutputDirectory = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);

      SetExtractionConfig(initialExtractionConfig, false);

      initialized = true;
    }

    public void ExtractOnShown(ExtractionConfig extractionConfig)
    {
      SetExtractionConfig(extractionConfig, true);

      doExtractOnShown = true;
    }

    protected override void OnShown(EventArgs e)
    {
      base.OnShown(e);

      if (doExtractOnShown) {
        IsEdittingExtractionConfigDirty = true;
        ExtractWithEdittingConfig();
      }
      else {
        edittingExtractionConfig.ClearTracks();
        IsEdittingExtractionConfigDirty = false;
      }
    }

    private void toolStripMenuItemFile_Quit_Click(object sender, EventArgs e)
    {
      Close();
    }

    protected override void OnClosing(CancelEventArgs e)
    {
      if (backgroundWorkerExtraction.IsBusy) {
        var result = MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_OnClosingAbortExtraction"), Application.ProductName, MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation, MessageBoxDefaultButton.Button2);

        if (result == DialogResult.Yes) {
          // abort
          backgroundWorkerExtraction.CancelAsync();

          // wait for RunWorkerCompleted
          while (backgroundWorkerExtraction.IsBusy) {
            Application.DoEvents();
            Thread.Sleep(250);
          }
        }
        else {
          // continue extraction
          e.Cancel = true;
        }
      }

      e.Cancel |= !CheckAndConfirmSaveExtractionConfig(true);

      base.OnClosing(e);
    }

    private void UpdateText()
    {
      var configFile = CurrentConfigFilePath ?? productSelector.SelectedProduct.Title;

      Text = string.Format("{0}{1} - {2}",
                           (IsEdittingExtractionConfigDirty || edittingExtractionConfig.AreAnyTracksDirty) ? "*" : string.Empty,
                           configFile,
                           ThbgmExtractor.GetText("ThbgmExtractor"));
    }

    private void SetSelectedEncoderConfigCheckedState()
    {
      var selectedEncoderConfig = edittingExtractionConfig.EncoderConfig ?? EncoderConfig.Empty;

      foreach (var item in toolStripMenuItemExtract_EncoderConfigs.DropDownItems) {
        if (item is EncoderConfigToolStripMenuItem) {
          var i = item as EncoderConfigToolStripMenuItem;

          i.Checked = (selectedEncoderConfig == i.EncoderConfig);
        }
      }
    }

    private void SetSelectedPlaylistFormatCheckedState()
    {
      var selectedPlaylistFormat = edittingExtractionConfig.PlaylistFormat;

      foreach (var item in toolStripMenuItemExtract_PlaylistFormats.DropDownItems) {
        if (item is PlaylistFormatToolStripMenuItem) {
          var i = item as PlaylistFormatToolStripMenuItem;

          i.Checked = (selectedPlaylistFormat == i.PlaylistFormat);
        }
      }
    }

    private void toolStripMenuItemTrackList_EditCollective_Click(object sender, EventArgs e)
    {
      Config.CollectiveEffectEditDefaults.OpenConfigAppliedDialog(delegate(CollectiveEffectEditDialog dialog) {
        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        foreach (var track in trackListView.Tracks) {
          track.Effect.FadingCurveType = dialog.FadingCurveType;
          track.Effect.FadingCurveCoefficient = dialog.FadingCurveCoefficient;

          switch (dialog.EffectLengthBasis) {
            case EffectLengthBasis.FadingStart:
              track.Effect.SetLength(dialog.FadingStart, dialog.FadingLength, dialog.TrailingSilenceLength);
              break;
            case EffectLengthBasis.TotalLength:
              track.Effect.SetTotalLength(dialog.TotalLength, dialog.FadingLength, dialog.TrailingSilenceLength);
              break;
            case EffectLengthBasis.TimesToRepeat:
              track.Effect.SetLength(track.TrackInfo, dialog.TimesToRepeat, dialog.FadingLength, dialog.FadingOffset, dialog.TrailingSilenceLength);
              break;
          }
        }
      });
    }

    private void toolStripMenuItemEdit_EditFileName_Click(object sender, EventArgs e)
    {
      Config.TrackFormatInfoDefaults.OpenConfigAppliedFileNameEditDialog(delegate(FileNameEditDialog dialog) {
        dialog.TrackFormatInfo = edittingExtractionConfig.TrackFormatInfo.Clone();
        dialog.Format = edittingExtractionConfig.TrackFormatString;

        if (trackListView.SelectedTrack == null) {
          if (edittingExtractionConfig.Product == null)
            dialog.TrackForExample = null;
          else
            dialog.TrackForExample = edittingExtractionConfig.Product.Tracks[0];
        }
        else {
          dialog.TrackForExample = trackListView.SelectedTrack.TrackInfo;
        }

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        edittingExtractionConfig.TrackFormatInfo = dialog.TrackFormatInfo.Clone();
        edittingExtractionConfig.TrackFormatString = dialog.Format;

        IsEdittingExtractionConfigDirty = true;

        trackListView.TrackFormatInfo = dialog.TrackFormatInfo.Clone();
        trackListView.Format = dialog.Format;
      });
    }

    private void toolStripMenuItemFile_RecentFiles_Clear_Click(object sender, EventArgs e)
    {
      Config.RecentFiles.ClearExtractionConfigFiles();
    }

    internal void FocusBgmSourcePath()
    {
      productSelector.Focus();
      productSelector.FocusBgmSourcePath();
    }

#region "open/save"
    private void toolStripMenuItemFile_Open_BmgSource_Click(object sender, EventArgs e)
    {
      productSelector.ShowChangeBmgSourcePathDialog();
    }

    private void toolStripMenuItemFile_Open_ExtractionConfig_Click(object sender, EventArgs e)
    {
      using (var dialog = new OpenFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
          new[] {ThbgmExtractor.GetText("ExtractionConfigFile"), "*" + ExtractionConfig.FileExtension},
          new[] {Texts.GetText("AllFiles"), "*.*"},
        });

        dialog.FilterIndex = 0;
        dialog.FileName = CurrentConfigFilePath;
        dialog.InitialDirectory = string.IsNullOrEmpty(CurrentConfigFilePath) ? Environment.CurrentDirectory : Path.GetDirectoryName(CurrentConfigFilePath);
        dialog.RestoreDirectory = true;

        if (dialog.ShowDialog(this) == DialogResult.OK)
          LoadExtractionConfig(dialog.FileName, true);
      }
    }

    private void toolStripMenuItemFile_SaveExtractionConfigAs_Click(object sender, EventArgs e)
    {
      SaveExtractionConfigAs();
    }

    private void toolStripMenuItemFile_SaveExtractionConfig_Click(object sender, EventArgs e)
    {
      SaveExtractionConfig();
    }

    private void toolStripMenuItemFile_RecentFiles_DropDownOpening(object sender, EventArgs e)
    {
      toolStripMenuItemFile_RecentFiles.DropDownItems.Clear();

      var entryExists = false;

      foreach (var file in Config.RecentFiles.GetExtractionConfigFiles()) {
        entryExists = true;

        toolStripMenuItemFile_RecentFiles.DropDownItems.Add(new ExtractionConfigFileStripMenuItem(file, delegate(object se, EventArgs eargs) {
          LoadExtractionConfig((se as ExtractionConfigFileStripMenuItem).File, true);
        }));
      }

      if (entryExists) {
        toolStripMenuItemFile_RecentFiles.DropDownItems.Add(toolStripMenuItemFile_RecentFiles_Separator);
        toolStripMenuItemFile_RecentFiles_Clear.Enabled = true;
      }
      else {
        toolStripMenuItemFile_RecentFiles_Clear.Enabled = false;
      }

      toolStripMenuItemFile_RecentFiles.DropDownItems.Add(toolStripMenuItemFile_RecentFiles_Clear);
    }
#endregion

#region "save/load extraction config"
    private bool CheckAndConfirmSaveExtractionConfig(bool checkAllConfig)
    {
      if (!initialized)
        return true;

      if (!(edittingExtractionConfig.AreAnyTracksDirty || (checkAllConfig && IsEdittingExtractionConfigDirty)))
        return true;

      var result = MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_CheckAndConfirmSaveExtractionConfig"), Application.ProductName, MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question, MessageBoxDefaultButton.Button1);

      switch (result) {
        case DialogResult.Yes:
          return SaveExtractionConfig() == DialogResult.OK;

        case DialogResult.No:
          return true;

        default:
          return false; // cancel
      }
    }

    public void LoadExtractionConfig(string path, bool persistent)
    {
      try {
        if (!CheckAndConfirmSaveExtractionConfig(true))
          return;

        SetExtractionConfig(ExtractionConfig.LoadFrom(path), true);

        if (persistent)
          CurrentConfigFilePath = path;

        IsEdittingExtractionConfigDirty = false;

        if (persistent)
          Config.RecentFiles.AddExtractionConfigFile(path);
      }
      catch (InvalidDataException ex) {
        MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_LoadExtractionConfig_InvalidDataException", ex.InnerException.Message), Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
        Console.Error.WriteLine(ex);
      }
      catch (NotSupportedException) {
        MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_LoadExtractionConfig_NotSupportedException"), Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
      }
    }

    private void SetExtractionConfig(ExtractionConfig newExtractionConfig, bool setTracks)
    {
      var propertyChanging = new PropertyChangingEventHandler(delegate(object sender, PropertyChangingEventArgs e) {
        var config = sender as ExtractionConfig;

        if (e.PropertyName == "BgmSourcePath") {
          lastBgmSourcePath = config.BgmSourcePath;
        }
        else if (e.PropertyName == "Product") {
          if (string.IsNullOrEmpty(productSelector.BgmSourcePath) || productSelector.BgmSourcePath == config.Product.InstalledOrDefaultBgmSourcePath)
            productSelector.BgmSourcePath = string.Empty;

          config.BgmSourcePath = lastBgmSourcePath;

          if (CheckAndConfirmSaveExtractionConfig(false)) {
            config.BgmSourcePath = productSelector.BgmSourcePath;

            // force save-as in next time
            CurrentConfigFilePath = null;

            return; // extraction config is not dirty or saved
          }
          else /* canceled */ {
            // revert bgm source and product
            productSelector.SelectedProduct = config.Product;
            productSelector.BgmSourcePath = config.BgmSourcePath;
            player.BgmSourcePath = config.BgmSourcePath;

            (e as CancellablePropertyChangingEventArgs).Cancel = true;
          }
        }
      });

      var propertyChanged = new PropertyChangedEventHandler(delegate(object sender, PropertyChangedEventArgs e) {
        var config = sender as ExtractionConfig;

        if (e.PropertyName == "Product" && string.IsNullOrEmpty(productSelector.BgmSourcePath))
          productSelector.BgmSourcePath = config.Product.InstalledOrDefaultBgmSourcePath;

        UpdateText();
      });

      edittingExtractionConfig.PropertyChanging -= propertyChanging;
      edittingExtractionConfig.PropertyChanged -= propertyChanged;

      productSelector.SelectedProduct = newExtractionConfig.Product;
      productSelector.BgmSourcePath = newExtractionConfig.BgmSourcePath;

      if (setTracks) {
        trackListView.SetProductAndTracks(newExtractionConfig.Product, newExtractionConfig.Tracks);
      }
      else {
        trackListView.Product = newExtractionConfig.Product;

        newExtractionConfig.Tracks = trackListView.Tracks;
      }

      trackListView.TrackFormatInfo = newExtractionConfig.TrackFormatInfo.Clone();
      trackListView.Format          = newExtractionConfig.TrackFormatString;

      toolStripMenuItemExtract_EncoderConfigs_SaveSourceAfterEncode.Checked = newExtractionConfig.SaveSourceAfterEncode;

      edittingExtractionConfig = newExtractionConfig;

      edittingExtractionConfig.PropertyChanging += propertyChanging;
      edittingExtractionConfig.PropertyChanged += propertyChanged;

      SetSelectedEncoderConfigCheckedState();
      SetSelectedPlaylistFormatCheckedState();
    }

    private DialogResult SaveExtractionConfig()
    {
      if (string.IsNullOrEmpty(CurrentConfigFilePath)) {
        return SaveExtractionConfigAs();
      }
      else {
        edittingExtractionConfig.Save(CurrentConfigFilePath);

        edittingExtractionConfig.ClearTracks();
        IsEdittingExtractionConfigDirty = false;

        return DialogResult.OK;
      }
    }

    private DialogResult SaveExtractionConfigAs()
    {
      using (var dialog = new SaveFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
          new[] {ThbgmExtractor.GetText("ExtractionConfigFile"), "*" + ExtractionConfig.FileExtension},
          new[] {Texts.GetText("AllFiles"), "*.*"},
        });

        var initialFilePath = string.IsNullOrEmpty(CurrentConfigFilePath)
          ? Path.Combine(Environment.CurrentDirectory, Smdn.IO.PathUtils.RemoveInvalidFileNameChars(productSelector.SelectedProduct.Title + ExtractionConfig.FileExtension))
          : CurrentConfigFilePath;

        dialog.FilterIndex = 0;
        dialog.FileName = initialFilePath;
        dialog.InitialDirectory = Path.GetDirectoryName(initialFilePath);
        dialog.RestoreDirectory = true;

        var result = dialog.ShowDialog(this);

        if (result == DialogResult.OK) {
          CurrentConfigFilePath = dialog.FileName;
          edittingExtractionConfig.Save(dialog.FileName);

          Config.RecentFiles.AddExtractionConfigFile(dialog.FileName);

          edittingExtractionConfig.ClearTracks();
          IsEdittingExtractionConfigDirty = false;
        }

        return result;
      }
    }
#endregion

#region "import/export"
    private class EmbeddedProductInfoToolStripMenuItem : ToolStripMenuItem {
      public ProductInfo Product {
        get; private set;
      }

      public EmbeddedProductInfoToolStripMenuItem(ProductInfo product, EventHandler onClick)
        : base(product.Title, null, onClick)
      {
        this.Product = product;
      }
    }

    private void toolStripMenuItemFile_ExportTitleFile_DropDownOpening(object sender, EventArgs e)
    {
      toolStripMenuItemFile_Export_TitleFiles.DropDownItems.Clear();

      foreach (var product in ThbgmExtractor.Products) {
        if (product.IsEmbedded)
          toolStripMenuItemFile_Export_TitleFiles.DropDownItems.Add(new EmbeddedProductInfoToolStripMenuItem(product, toolStripMenuItemFile_Export_TitleFile_EmbeddedProductInfo_Click));
      }
    }

    private void toolStripMenuItemFile_Export_TitleFile_EmbeddedProductInfo_Click(object sender, EventArgs e)
    {
      var menuItem = sender as EmbeddedProductInfoToolStripMenuItem;

      if (menuItem == null)
        return;

      ExportTitleFile(menuItem.Product);
    }

    private void toolStripMenuItemFile_Export_THxxBGMFadeoutSettings_Click(object sender, EventArgs e)
    {
      ExportTHxxBGMFadeoutSettings();
    }
    
    private void toolStripMenuItemFile_Import_ThplayApplicationConfig_Click(object sender, EventArgs e)
    {
      ImportThplayApplicationConfig();
    }

    private void toolStripMenuItemFile_Import_THxxBGMFadeoutSettings_Click(object sender, EventArgs e)
    {
      ImportTHxxBGMFadeoutSettings();
    }

    private void toolStripMenuItemFile_Import_THxxBGMSettings_Click(object sender, EventArgs e)
    {
      ImportTHxxBGMSettings();
    }

    private void ExportTitleFile(ProductInfo product)
    {
      using (var dialog = new SaveFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
          new[] {Texts.GetText("ProductInfoFile"), "*.txt"},
          new[] {Texts.GetText("AllFiles"), "*.*"},
        });

        dialog.FilterIndex = 0;
        dialog.FileName = Path.Combine(Environment.CurrentDirectory, string.Format("titles_{0}.txt", product.Prefix));
        dialog.InitialDirectory = Environment.CurrentDirectory;
        dialog.RestoreDirectory = true;

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        product.Save(dialog.FileName);
      }
    }

    private void ExportTHxxBGMFadeoutSettings()
    {
      if (Guid.Empty.Equals(edittingExtractionConfig.Product.THxxBGMProductGuid)) {
        MessageBox.Show(this,
                        ThbgmExtractor.GetText("MessageBox_ExportTHxxBGMFadeoutSettings_THxxBGMUnsupportedProduct", edittingExtractionConfig.Product.Title),
                        Application.ProductName,
                        MessageBoxButtons.OK,
                        MessageBoxIcon.Error);
        return;
      }

      using (var dialog = new SaveFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
          new[] {ThbgmExtractor.GetText("THxxBGMFadeoutSettingsFile"), "*" + ExtractionConfig.THxxBGMFadeoutSettingsFileExtension},
          new[] {Texts.GetText("AllFiles"), "*.*"},
        });

        dialog.FilterIndex = 0;
        dialog.FileName = Path.Combine(Environment.CurrentDirectory, string.Format("{0}{1}", edittingExtractionConfig.Product.THxxBGMProductGuid.ToString("D").ToUpperInvariant(), ExtractionConfig.THxxBGMFadeoutSettingsFileExtension));
        dialog.InitialDirectory = Environment.CurrentDirectory;
        dialog.RestoreDirectory = true;

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        try {
          edittingExtractionConfig.SaveAsTHxxBGMFadeoutSettings(dialog.FileName);

          MessageBox.Show(this,
                          ThbgmExtractor.GetText("MessageBox_ExportTHxxBGMFadeoutSettings_ExportDone", dialog.FileName),
                          Application.ProductName,
                          MessageBoxButtons.OK,
                          MessageBoxIcon.Information);
        }
        catch (Exception ex) {
          Console.Error.WriteLine(ex);

          MessageBox.Show(this,
                          ThbgmExtractor.GetText("MessageBox_ExportTHxxBGMFadeoutSettings_ExportFailed"),
                          Application.ProductName,
                          MessageBoxButtons.OK,
                          MessageBoxIcon.Error);
        }
      }
    }

    private void ImportThplayApplicationConfig()
    {
      using (var dialog = new OpenFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
          new[] {ThbgmExtractor.GetText("ThplayApplicationConfigFile"), Thplay.ApplicationConfigFileName},
          new[] {Texts.GetText("AllFiles"), "*.*"},
        });

        dialog.FilterIndex = 0;
        dialog.FileName = Thplay.ApplicationConfigFileName;
        dialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
        dialog.RestoreDirectory = true;

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        try {
          Thplay.ImportApplicationConfig(dialog.FileName);

          MessageBox.Show(this,
                          ThbgmExtractor.GetText("MessageBox_ImportThplayApplicationConfig_ImportDone", dialog.FileName),
                          Application.ProductName,
                          MessageBoxButtons.OK,
                          MessageBoxIcon.Information);
        }
        catch (Exception ex) {
          Console.Error.WriteLine(ex);

          MessageBox.Show(this,
                          ThbgmExtractor.GetText("MessageBox_ImportThplayApplicationConfig_ImportFailed"),
                          Application.ProductName,
                          MessageBoxButtons.OK,
                          MessageBoxIcon.Error);
        }
      }
    }

    private void ImportTHxxBGMFadeoutSettings()
    {
      using (var dialog = new OpenFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
          new[] {ThbgmExtractor.GetText("THxxBGMFadeoutSettingsFile"), "*" + ExtractionConfig.THxxBGMFadeoutSettingsFileExtension},
          new[] {Texts.GetText("AllFiles"), "*.*"},
        });

        dialog.FilterIndex = 0;
        dialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
        dialog.RestoreDirectory = true;

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        try {
          LoadExtractionConfig(dialog.FileName, false);

          MessageBox.Show(this,
                          ThbgmExtractor.GetText("MessageBox_ImportTHxxBGMFadeoutSettings_ImportDone", dialog.FileName),
                          Application.ProductName,
                          MessageBoxButtons.OK,
                          MessageBoxIcon.Information);
        }
        catch (Exception ex) {
          Console.Error.WriteLine(ex);

          MessageBox.Show(this,
                          ThbgmExtractor.GetText("MessageBox_ImportTHxxBGMFadeoutSettings_ImportFailed"),
                          Application.ProductName,
                          MessageBoxButtons.OK,
                          MessageBoxIcon.Error);
        }
      }
    }

    private void ImportTHxxBGMSettings()
    {
      using (var dialog = new OpenFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
          new[] {ThbgmExtractor.GetText("THxxBGMSettingsFile"), THxxBGM.THxxBGMSettingsFileName},
          new[] {Texts.GetText("AllFiles"), "*.*"},
        });

        dialog.FilterIndex = 0;
        dialog.FileName = THxxBGM.THxxBGMSettingsFileName;
        dialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
        dialog.RestoreDirectory = true;

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        try {
          THxxBGM.ImportTHxxBGMSettings(dialog.FileName);

          MessageBox.Show(this,
                          ThbgmExtractor.GetText("MessageBox_ImportTHxxBGMSettings_ImportDone", dialog.FileName),
                          Application.ProductName,
                          MessageBoxButtons.OK,
                          MessageBoxIcon.Information);
        }
        catch (Exception ex) {
          Console.Error.WriteLine(ex);

          MessageBox.Show(this,
                          ThbgmExtractor.GetText("MessageBox_ImportTHxxBGMSettings_ImportFailed"),
                          Application.ProductName,
                          MessageBoxButtons.OK,
                          MessageBoxIcon.Error);
        }
      }
    }
#endregion

#region "drag'n drop"
    protected override void OnDragEnter(DragEventArgs drgevent)
    {
      base.OnDragEnter(drgevent);

      if (drgevent.Data.GetDataPresent(DataFormats.FileDrop))
        drgevent.Effect = DragDropEffects.All;
      else
        drgevent.Effect = DragDropEffects.None;
    }

    protected override void OnDragDrop(DragEventArgs drgevent)
    {
      try {
        var files = (string[])drgevent.Data.GetData(DataFormats.FileDrop, false);

        if (0 < files.Length) {
          if (Smdn.IO.PathUtils.AreExtensionEqual(files[0], ExtractionConfig.FileExtension))
            LoadExtractionConfig(files[0], true);
          else
            productSelector.SetBgmSourcePath(files[0], true);
        }
      }
      catch {
        // ignore exceptions
      }
      finally {
        base.OnDragDrop(drgevent);
      }
    }
#endregion

#region "begin extraction"
    private void toolStripMenuItemExtract_ExtractTrack_Click(object sender, EventArgs e)
    {
      ExtractSelectedTrack();
    }

    private void toolStripMenuItemTrackList_ExtractTrack_Click(object sender, EventArgs e)
    {
      ExtractSelectedTrack();
    }

    private void ExtractSelectedTrack()
    {
      var track = trackListView.SelectedTrack;

      if (track == null)
        return;

      var extractTrackConfig = edittingExtractionConfig.Clone();

      extractTrackConfig.Tracks = new Track[] {track};

      ConfigureAndExtract(extractTrackConfig, true);
    }

    private void toolStripMenuItemExtract_Extract_Click(object sender, EventArgs e)
    {
      ExtractWithEdittingConfig();
    }

    private void ExtractWithEdittingConfig()
    {
      ConfigureAndExtract(edittingExtractionConfig, false);
    }

    private void ConfigureAndExtract(ExtractionConfig baseConfig, bool extractSingleTrack)
    {
      if (backgroundWorkerExtraction.IsBusy) {
        MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_ExtractionProceeding"), Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
        return;
      }

      var configurator = new ExtractionPreconfigurator(this, baseConfig);

      configurator.Config.PropertyChanged += delegate(object sender, PropertyChangedEventArgs e) {
        if (e.PropertyName == "OutputDirectory")
          edittingExtractionConfig.OutputDirectory = (sender as ExtractionConfig).OutputDirectory;
      };

      if (!configurator.Configure(extractSingleTrack))
        return; // canceled

      var config = configurator.Config;

      progressBarExtract.Minimum = 0;
      progressBarExtract.Maximum = config.Tracks.Count() * 100;

      progressBarEncode.Minimum = 0;
      progressBarEncode.Maximum = config.Tracks.Count();

      splitContainerExtraction.Panel2Collapsed = false;

      backgroundWorkerExtraction.RunWorkerAsync(config);
    }

    private void buttonAbortExtraction_Click(object sender, EventArgs e)
    {
      // abort
      if (!backgroundWorkerExtraction.IsBusy)
        return;

      backgroundWorkerExtraction.CancelAsync();
    }
#endregion

#region "backgroundWorkerExtraction"
    private void backgroundWorkerExtraction_DoWork(object sender, DoWorkEventArgs e)
    {
      var config = e.Argument as ExtractionConfig;
      var context = Extractor.StartExtract(config);

      backgroundWorkerExtraction.ReportProgress(0, context);

      for (;;) {
        if (!context.IsAbortRequested && backgroundWorkerExtraction.CancellationPending) {
          e.Cancel = true;
          Extractor.Abort(context);
        }

        if (Extractor.WaitForExtractionFinished(context, 500))
          break;
        else
          backgroundWorkerExtraction.ReportProgress(0, context);
      }

      backgroundWorkerExtraction.ReportProgress(0, context);

      e.Result = context;
    }
    
    private void backgroundWorkerExtraction_ProgressChanged(object sender, ProgressChangedEventArgs e)
    {
      var context = e.UserState as ExtractionContext;
      var progress = context.Progress;

      labelExtractProgress.Text = ThbgmExtractor.GetText("labelExtractProgress_Text", progress.ExtractedTrackCount, progress.TotalTrackCount);

      if (EncoderConfig.IsNullOrEmpty(context.Config.EncoderConfig))
        labelEncodeProgress.Text = string.Empty;
      else
        labelEncodeProgress.Text = ThbgmExtractor.GetText("labelEncodeProgress_Text", progress.EncodedTrackCount, progress.TotalTrackCount);

      labelProgressStatus.Text = progress.Status;

      var perTrackProgressPercentage = (progress.TrackLength == 0L || progress.ExtractedTrackCount == progress.TotalTrackCount)
        ? 0
        : (int)(100L * progress.ExtractedLength / progress.TrackLength);

      progressBarExtract.Value = progress.ExtractedTrackCount * 100 + perTrackProgressPercentage;
      progressBarEncode.Value = progress.EncodedTrackCount;
    }

    private void backgroundWorkerExtraction_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
    {
      progressBarExtract.Maximum = 1;
      progressBarExtract.Value = 1;

      progressBarEncode.Maximum = 1;
      progressBarEncode.Value = 1;

      labelExtractProgress.Text = string.Empty;
      labelEncodeProgress.Text = string.Empty;

      if (e.Error != null) {
        Console.Error.WriteLine(e.Error);

        MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_ExtractionFinished_AbortedByError"), Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Information, MessageBoxDefaultButton.Button1);
      }
      else if (e.Cancelled) {
        MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_ExtractionFinished_Aborted"), Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Information, MessageBoxDefaultButton.Button1);
      }
      else {
        var context = e.Result as ExtractionContext;
        var result = MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_ExtractionFinished_ConfirmOpenOutputDirectory"), Application.ProductName, MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button1);

        if (result == DialogResult.Yes) {
          try {
            FileManager.Browse(context.Config.OutputDirectory);
          }
          catch {
            // ignore exceptions
          }
        }
      }

      splitContainerExtraction.Panel2Collapsed = true;
    }
#endregion

    private void timerLog_Tick(object sender, EventArgs e)
    {
      if (textBoxLog.TextLength == ThbgmExtractor.Log.Length)
        return;

      textBoxLog.Text = ThbgmExtractor.Log.ToString();
      textBoxLog.Select(textBoxLog.TextLength - 1, 0);
      textBoxLog.ScrollToCaret();
    }

    private bool initialized = false;
    private ExtractionConfig edittingExtractionConfig = new ExtractionConfig();
    private bool isEdittingExtractionConfigDirty = false;
    private string currentConfigFilePath = null;
    private string lastBgmSourcePath = string.Empty;
    private bool doExtractOnShown = false;
    private List<ToolStripItem> editToolStripItems = new List<ToolStripItem>();
    private List<ToolStripItem> trackListToolStripItems = new List<ToolStripItem>();
    private ToolStripSeparator contextMenuDummyItem = new ToolStripSeparator();
  }
}
