using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Resources;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;

using Smdn.Formats.Thbgm;
using Smdn.IO;
using Smdn.Windows.Forms.ThbgmPlayer;

namespace Smdn.Applications.ThbgmExtractor {
  public static class ThbgmExtractor {
    private static StringBuilder log = new StringBuilder(1024 * 1024);

    public static StringBuilder Log {
      get { return log; }
    }

    private static ResourceManager resourceManager = new ResourceManager("Smdn.Applications.ThbgmExtractor.Texts", typeof(ThbgmExtractor).Assembly);
    private static CultureInfo resourceCulture = null;

    public static System.Globalization.CultureInfo Culture {
      get { return resourceCulture; }
      set { resourceCulture = value; }
    }

    private static List<ProductInfo> products = new List<ProductInfo>();

    public static IEnumerable<ProductInfo> Products {
      get { return products; }
    }

    public static IEnumerable<EncoderConfig> EncoderConfigs {
      get; private set;
    }

    public static string VersionString {
      get { return string.Format("{0} {1}", Application.ProductName, Application.ProductVersion); }
    }

    [STAThread]
    public static void Main(string[] args)
    {
#if DEBUG
      resourceCulture = System.Globalization.CultureInfo.GetCultureInfo("en");
      Texts.Culture = resourceCulture;
      System.Threading.Thread.CurrentThread.CurrentUICulture = resourceCulture;
#endif

      var logWriter = TextWriter.Synchronized(new StringWriter(log));

#if DEBUG
      Console.SetError(logWriter);

      Console.WriteLine("current directory: {0}", Environment.CurrentDirectory);
      Console.WriteLine("executable directory: {0}", Path.GetDirectoryName(Application.ExecutablePath));
#else
      Console.SetOut(logWriter);
      Console.SetError(logWriter);
#endif

      products.AddRange(ProductInfo.EmbeddedProducts);

#if DEBUG
      var searchDirectory = Path.GetDirectoryName(Application.ExecutablePath);
#else
      var searchDirectory = Path.GetDirectoryName(Application.ExecutablePath);
#endif

      LoadProductInfo(searchDirectory);

      Config.Load(searchDirectory, products);

      LoadEncoderConfig(searchDirectory);

      Application.EnableVisualStyles();
      Application.DoEvents();

      using (var dialog = new MainDialog()) {
        if (0 < args.Length) {
          if (PathUtils.AreExtensionEqual(args[0], ExtractionConfig.FileExtension)) {
            dialog.LoadExtractionConfig(args[0], true);
          }
          else {
            var product = ProductInfo.FindMatchedProduct(args[0], products);

            if (product != null) {
              var config = new ExtractionConfig();

              config.Product = product;
              config.BgmSourcePath = product.InstalledBgmSourcePath;
              config.EncoderConfig = EncoderConfig.Empty;
              config.Tracks = (new List<TrackInfo>(product.Tracks)).ConvertAll(delegate(TrackInfo trackInfo) {
                return new Track(trackInfo);
              });

              dialog.ExtractOnShown(config);
            }
          }
        }

        Application.Run(dialog);
      }

      Config.Save(products);
    }

    private static void LoadProductInfo(string searchDirectory)
    {
      try {
        var directory = new DirectoryInfo(searchDirectory);
        var titleFileRegex = new Regex(@"titles.*\.txt$", RegexOptions.Singleline | RegexOptions.IgnoreCase);

        foreach (var file in directory.GetFiles(SearchOption.TopDirectoryOnly, delegate(FileInfo file) {
          return titleFileRegex.IsMatch(file.Name);
        })) {
          try {
            using (var stream = file.OpenRead()) {
              var product = ProductInfo.LoadFrom(stream);

              products.Add(product);

              Console.WriteLine("loaded: {0} ({0})", file, product.Title);
            }
          }
          catch (Exception ex) {
            Console.Error.WriteLine("file: {0}{1}{2}", file, Environment.NewLine, ex);
          }
        }
      }
      catch (Exception ex) {
        Console.Error.WriteLine(ex);
      }
    }

    private static void LoadEncoderConfig(string searchDirectory)
    {
      try {
        EncoderConfigs = EncoderConfig.LoadFrom(Path.Combine(searchDirectory, EncoderConfig.ConfigFileName));
      }
      catch (Exception ex) {
        EncoderConfigs = new List<EncoderConfig>(new[] {EncoderConfig.Empty});
        Console.Error.WriteLine(ex);
      }
    }

    public static string GetText(string name, params object[] args)
    {
      return string.Format(resourceManager.GetString(name, resourceCulture) ?? string.Empty, args);
    }
  }
}
